"""Worker functions for regridding."""

import logging
from operator import itemgetter

import numpy as np

from .. import _found_ESMF
from ..data import Data
from ..dimensioncoordinate import DimensionCoordinate
from ..functions import regrid_logging

if _found_ESMF:
    try:
        import ESMF
    except Exception:
        _found_ESMF = False

from .regridoperator import (
    RegridOperator,
    conservative_regridding_methods,
    regrid_method_map,
    regridding_methods,
)

logger = logging.getLogger(__name__)


def regrid_compute_mass_grid(
    valuefield,
    areafield,
    dofrac=False,
    fracfield=None,
    uninitval=422397696.0,
):
    """Compute the mass of an `ESMF` Field.

    :Parameters:

        valuefield: `ESMF.Field`
            This contains data values of a field built on the cells of
            a grid.

        areafield: `ESMF.Field`
            This contains the areas associated with the grid cells.

        fracfield: `ESMF.Field`
            This contains the fractions of each cell which contributed
            to a regridding operation involving 'valuefield.

        dofrac: `bool`
            This gives the option to not use the 'fracfield'.

        uninitval: `float`
            The value uninitialised cells take.

    :Returns:

        `float`
            The mass of the data field.

    """
    mass = 0.0
    areafield.get_area()

    ind = np.where(valuefield.data != uninitval)

    if dofrac:
        mass = np.sum(
            areafield.data[ind] * valuefield.data[ind] * fracfield.data[ind]
        )
    else:
        mass = np.sum(areafield.data[ind] * valuefield.data[ind])

    return mass


def regrid_get_latlon(f, name, method, axes=None):
    """Get latitude and longitude coordinate information.

    Retrieve the latitude and longitude coordinates of a field, as
    well as some associated information. If 1-d lat/lon coordinates
    are found then these are returned. Otherwise if 2-d lat/lon
    coordinates found then these are returned.

    :Parameters:

        f: `Field`
            The source or destination field from which to get the
            information.

        name: `str`
            A name to identify the field in error messages. Either
            ``'source'`` or ``'destination'``.

        method: `str`
            The regridding method.

        axes: `dict`, optional
            A dictionary specifying the X and Y axes, with keys
            ``'X'`` and ``'Y'``.

            *Parameter example:*
              ``axes={'X': 'ncdim%x', 'Y': 'ncdim%y'}``

            *Parameter example:*
              ``axes={'X': 1, 'Y': 0}``

    :Returns:

        `list`, `list`, `list`, `list`, `bool`
            * The keys of the X and Y dimension coordinates.

            * The sizes of the X and Y dimension coordinates.

            * The keys of the X and Y coordinate (1-d dimension
              coordinate, or 2-d auxilliary coordinates).

            * The X and Y coordinates (1-d dimension coordinates or
              2-d auxilliary coordinates).

            * True if 2-d auxiliary coordinates are returned or if 1-d
              X and Y coordinates are returned, which are not
              lon/lat.

    """
    data_axes = f.constructs.data_axes()

    if axes is None:
        # Retrieve the field construct's X and Y dimension
        # coordinates
        x_key, x = f.dimension_coordinate(
            "X",
            item=True,
            default=ValueError(
                f"No unique X dimension coordinate found for the {name} "
                f"field {f!r}. If none is present you "
                "may need to specify the axes keyword."
            ),
        )
        y_key, y = f.dimension_coordinate(
            "Y",
            item=True,
            default=ValueError(
                f"No unique Y dimension coordinate found for the {name} "
                f"field {f!r}. If none is present you "
                "may need to specify the axes keyword."
            ),
        )

        x_axis = data_axes[x_key][0]
        y_axis = data_axes[y_key][0]

        x_size = x.size
        y_size = y.size
    else:
        # --------------------------------------------------------
        # Axes have been provided
        # --------------------------------------------------------
        for key in ("X", "Y"):
            if key not in axes:
                raise ValueError(
                    f"Key {key!r} must be specified for axes of {name} "
                    f"field {f!r}."
                )

        if axes["X"] in (1, 0) and axes["Y"] in (0, 1):
            # Axes specified by integer position in dimensions of
            # lat and lon 2-d auxiliary coordinates
            if axes["X"] == axes["Y"]:
                raise ValueError(
                    "The X and Y axes must be distinct, but they are the same "
                    "for {name} field {f!r}."
                )

            lon_key, lon = f.auxiliary_coordinate(
                "X", item=True, filter_by_naxes=(2,), default=(None, None)
            )
            lat_key, lat = f.auxiliary_coordinate(
                "Y", item=True, filter_by_naxes=(2,), default=(None, None)
            )
            if lon is None:
                raise ValueError(
                    "The X axis does not correspond to a longitude coordinate "
                    f"for {name} field {f!r}."
                )
            if lat is None:
                raise ValueError(
                    "The Y axis does not correspond to a latitude coordinate "
                    f"for {name} field {f!r}."
                )

            if lat.shape != lon.shape:
                raise ValueError(
                    "The shape of the latitude and longitude coordinates "
                    "must be equal but they differ for {name} field {f!r}."
                )

            lon_axes = data_axes[lon_key]
            lat_axes = data_axes[lat_key]
            if lat_axes != lon_axes:
                raise ValueError(
                    "The domain axis constructs spanned by the latitude and "
                    "longitude coordinates should be the same, but they "
                    "differ for {name} field {f!r}."
                )

            x_axis = lon_axes[axes["X"]]
            y_axis = lat_axes[axes["Y"]]
        else:
            x_axis = f.domain_axis(
                axes["X"],
                key=True,
                default=ValueError(
                    f"'X' axis specified for {name} {f!r} field not found."
                ),
            )

            y_axis = f.domain_axis(
                axes["Y"],
                key=True,
                default=ValueError(
                    f"'Y' axis specified for {name} field {f!r} not found."
                ),
            )

        domain_axes = f.domain_axes(todict=True)
        x_size = domain_axes[x_axis].get_size()
        y_size = domain_axes[y_axis].get_size()

    axis_keys = [x_axis, y_axis]
    axis_sizes = [x_size, y_size]

    # If 1-d latitude and longitude coordinates for the field are
    # not found search for 2-d auxiliary coordinates.
    if axes is not None or not x.Units.islongitude or not y.Units.islatitude:
        lon_found = False
        lat_found = False

        for key, aux in f.auxiliary_coordinates(
            filter_by_naxes=(2,), todict=True
        ).items():
            if aux.Units.islongitude:
                if lon_found:
                    raise ValueError(
                        "The 2-d auxiliary longitude coordinate "
                        f"of the {name} field {f!r} is not unique."
                    )
                else:
                    lon_found = True
                    x = aux
                    x_key = key

            if aux.Units.islatitude:
                if lat_found:
                    raise ValueError(
                        "The 2-d auxiliary latitude coordinate "
                        f"of the {name} field {f!r} is not unique."
                    )
                else:
                    lat_found = True
                    y = aux
                    y_key = key

        if not lon_found or not lat_found:
            raise ValueError(
                "Both longitude and latitude coordinates "
                f"were not found for the {name} field {f!r}."
            )

        if axes is not None:
            if set(axis_keys) != set(data_axes[x_key]):
                raise ValueError(
                    "Axes of longitude do not match "
                    f"those specified for {name} field {f!r}."
                )

            if set(axis_keys) != set(data_axes[y_key]):
                raise ValueError(
                    "Axes of latitude do not match "
                    f"those specified for {name} field {f!r}."
                )

        coords_2D = True
    else:
        coords_2D = False

    # Check for size 1 latitude or longitude dimensions if source grid
    # (a size 1 dimension is only problematic for the source grid in ESMF)
    if (
        name == "source"
        and method in ("linear", "bilinear", "patch")
        and (x_size == 1 or y_size == 1)
    ):
        raise ValueError(
            f"Neither the longitude nor latitude dimensions of the {name}"
            f"field {f!r} can be of size 1 for {method!r} regridding."
        )

    coord_keys = [x_key, y_key]
    coords = [x, y]

    return axis_keys, axis_sizes, coord_keys, coords, coords_2D


def get_cartesian_coords(f, name, axes):
    """Retrieve the specified cartesian dimension coordinates of the
    field and their corresponding keys.

    :Parameters:

        f: `Field`
           The field from which to get the coordinates.

        name: `str`
            A name to identify the field in error messages.

        axes: sequence of `str`
            Specifiers for the dimension coordinates to be
            retrieved. See `cf.Field.domain_axes` for details.

    :Returns:

        `list`, `list`
            A list of the keys of the dimension coordinates; and a
            list of the dimension coordinates retrieved.

    """
    axis_keys = []
    for axis in axes:
        key = f.domain_axis(axis, key=True)
        axis_keys.append(key)

    coords = []
    for key in axis_keys:
        d = f.dimension_coordinate(filter_by_axis=(key,), default=None)
        if d is None:
            raise ValueError(
                f"No unique {name} dimension coordinate "
                f"matches key {key!r}."
            )

        coords.append(d.copy())

    return axis_keys, coords


def regrid_get_axis_indices(f, axis_keys):
    """Get axis indices and their orders in rank of this field.

    :Parameters:

         f: `Field`
            The source or destination field. This field might get
            size-1 dimensions inserted into its data in-place.

        axis_keys: sequence
            A sequence of domain axis identifiers for the axes being
            regridded.

    :Returns:

        `list`, `list`, `numpy.ndarray`
            * The indices of the specified axes.

            * The indices of the non-specified field data axes .

            * The rank order of the axes.

    """
    data_axes = f.get_data_axes()

    # Get the positions of the axes
    axis_indices = []
    for axis_key in axis_keys:
        try:
            axis_index = data_axes.index(axis_key)
        except ValueError:
            f.insert_dimension(axis_key, position=0, inplace=True)
            axis_index = 0

        axis_indices.append(axis_index)

    # Get the rank order of the positions of the axes
    tmp = np.array(axis_indices)
    tmp = tmp.argsort()
    n = len(tmp)
    order = np.empty((n,), dtype=int)
    order[tmp] = np.arange(n)

    non_regridding_axis_indices = [
        i for i in range(f.ndim) if i not in axis_indices
    ]

    return axis_indices, non_regridding_axis_indices, order


def regrid_get_coord_order(f, axis_keys, coord_keys):
    """Get the ordering of the axes for each N-d auxiliary coordinate.

    :Parameters:

         f: `Field`
            The source or destination field.

        axis_keys: sequence
            A sequence of axis keys.

        coord_keys: sequence
            A sequence of keys for each of the N-d auxiliary
            coordinates.

    :Returns:

        `list`
            A list of lists specifying the ordering of the axes for
            each N-d auxiliary coordinate.

    """
    coord_axes = [f.get_data_axes(coord_key) for coord_key in coord_keys]
    coord_order = [
        [coord_axis.index(axis_key) for axis_key in axis_keys]
        for coord_axis in coord_axes
    ]

    return coord_order


def regrid_get_section_shape(src, axis_sizes, axis_indices):
    """Get the shape of each regridded section.

    :Parameters:

        src: `Field`
            The source field.

        axis_sizes: sequence
            A sequence of the sizes of each axis along which the
            section.  will be taken

        axis_indices: sequence
            A sequence of the same length giving the axis index of
            each axis.

    :Returns:

        `list`
            A list of integers defining the shape of each section.

    """
    shape = [1] * src.ndim
    for i, axis_index in enumerate(axis_indices):
        shape[axis_index] = axis_sizes[i]

    return shape


def regrid_check_bounds(src_coords, dst_coords, method, ext_coords=None):
    """Check the bounds of the coordinates for regridding.

    :Parameters:

        src_coords: sequence
            A sequence of the source coordinates.

        dst_coords: sequence
            A sequence of the destination coordinates.

        method: `str`
            A string indicating the regrid method.

        ext_coords: `None` or sequence
            If a sequence of extension coordinates is present these
            are also checked. Only used for cartesian regridding when
            regridding only 1 (only 1!) dimension of an N>2
            dimensional field. In this case we need to provided the
            coordinates of the dimensions that aren't being regridded
            (that are the same in both src and dst grids) so that we
            can create a sensible ESMF grid object.

    :Returns:

        `None`

    """
    if method not in conservative_regridding_methods:
        # Bounds are only not needed for conservative methods
        return

    for name, coords in zip(
        ("Source", "Destination"), (src_coords, dst_coords)
    ):
        for coord in coords:
            if not coord.has_bounds():
                raise ValueError(
                    f"{name} {coord!r} coordinates must have bounds "
                    "for conservative regridding."
                )

            if not coord.contiguous(overlap=False):
                raise ValueError(
                    f"{name} {coord!r} coordinates must have "
                    "contiguous, non-overlapping bounds "
                    "for conservative regridding."
                )

    if ext_coords is not None:
        for coord in ext_coords:
            if not coord.has_bounds():
                raise ValueError(
                    f"{coord!r} dimension coordinates must have "
                    "bounds for conservative regridding."
                )
            if not coord.contiguous(overlap=False):
                raise ValueError(
                    f"{coord!r} dimension coordinates must have "
                    "contiguous, non-overlapping bounds "
                    "for conservative regridding."
                )


def regrid_check_method(method):
    """Check that the regrid method is valid.

    If it is not valid then an exception is raised.

    :Parameters:

        method: `str`
            The regridding method.

    :Returns:

        `None`

    """
    if method not in regridding_methods:
        raise ValueError(
            "Can't regrid: Must set a valid regridding method from "
            f"{regridding_methods}. Got: {method!r}"
        )
    elif method == "bilinear":
        logger.info(
            "Note the 'bilinear' method argument has been renamed to "
            "'linear' at version 3.2.0. It is still supported for now "
            "but please use 'linear' in future. "
            "'bilinear' will be removed at version 4.0.0."
        )


def regrid_check_use_src_mask(use_src_mask, method):
    """Check the setting of the use_src_mask parameter.

    An exception is raised is the setting is incorrect relative to the
    regridding method.

    :Parameters:

        use_src_mask: `bool`
            Whether to use the source mask in regridding.

        method: `str`
            The regridding method.

    :Returns:

        `None`

    """
    if not use_src_mask and not method == "nearest_stod":
        raise ValueError(
            "The use_src_mask parameter can only be False when using the "
            "'nearest_stod' regridding method."
        )


def regrid_get_reordered_sections(
    src, axis_order, regrid_axes, regrid_axis_indices
):
    """Get a dictionary of the data sections for regridding and a list
    of its keys reordered if necessary so that they will be looped over
    in the order specified in axis_order.

    :Parameters:

        src: `Field`
            The source field.

        axis_order: `None` or sequence of axes specifiers.
            If `None` then the sections keys will not be reordered. If
            a particular axis is one of the regridding axes or is not
            found then a ValueError will be raised.

        regrid_axes: sequence
            A sequence of the keys of the regridding axes.

        regrid_axis_indices: sequence
            A sequence of the indices of the regridding axes.

    :Returns:

        `list`, `dict`
            An ordered list of the section keys; and a dictionary of
            the data sections for regridding.

    """
    # If we had dynamic masking, we wouldn't need this method, we
    # could simply replace it in the calling function with a call to
    # Data.section. However, we don't have it, so this allows us to
    # possibly reduce the number of transitions between different
    # masks - each change is slow.
    data_axes = src.get_data_axes()

    axis_indices = []
    if axis_order is not None:
        for axis in axis_order:
            axis_key = src.dimension_coordinate(
                filter_by_axis=(axis,),
                default=None,
                key=True,
            )
            if axis_key is not None:
                if axis_key in regrid_axes:
                    raise ValueError("Cannot loop over regridding axes.")

                try:
                    axis_indices.append(data_axes.index(axis_key))
                except ValueError:
                    # The axis has been squeezed so do nothing
                    pass

            else:
                raise ValueError(f"Source field axis not found: {axis!r}")

    # Section the data
    sections = src.data.section(regrid_axis_indices)

    # Reorder keys correspondingly if required
    if axis_indices:
        section_keys = sorted(sections.keys(), key=itemgetter(*axis_indices))
    else:
        section_keys = sections.keys()

    return section_keys, sections


def regrid_get_destination_mask(
    f, dst_order, axes=("X", "Y"), cartesian=False, coords_ext=None
):
    """Get the mask of the destination field.

    :Parameters:

        f: `Field`
            The destination field.

        dst_order: sequence, optional
            The order of the destination axes.

        axes: optional
            The axes the data is to be sectioned along.

        cartesian: `bool`, optional
            Whether the regridding is Cartesian or spherical.

        coords_ext: sequence, optional
            In the case of Cartesian regridding, extension coordinates
            (see _regrid_check_bounds for details).

    :Returns:

        `numpy.ndarray`
            The mask.

    """
    data_axes = f.get_data_axes()

    indices = {axis: [0] for axis in data_axes if axis not in axes}

    g = f.subspace(**indices)
    g = g.squeeze(tuple(indices)).transpose(dst_order)

    dst_mask = g.mask.array

    if cartesian:
        tmp = []
        for coord in coords_ext:
            tmp.append(coord.size)
            dst_mask = np.tile(dst_mask, tmp + [1] * dst_mask.ndim)

    return dst_mask


def regrid_fill_fields(src_data, srcfield, dstfield, fill_value):
    """Fill the source and destination Fields.

    Fill the source Field with data and the destination Field with
    fill values.

    :Parameters:

        src_data: ndarray
            The data to fill the source field with.

        srcfield: `ESMF.Field`
            The source field.

        dstfield: `ESMF.Field`
            The destination field. This get always gets initialised with
            missing values.

        fill_value:
            The fill value with which to fill *dstfield*

    :Returns:

        `None`

    """
    srcfield.data[...] = np.ma.MaskedArray(src_data, copy=False).filled(
        fill_value
    )
    dstfield.data[...] = fill_value


def regrid_compute_field_mass(
    f,
    _compute_field_mass,
    k,
    srcgrid,
    srcfield,
    srcfracfield,
    dstgrid,
    dstfield,
):
    """Compute the field mass for conservative regridding.

    The mass should be the same before and after regridding.

    :Parameters:

        f: `Field`
            The source field.

        _compute_field_mass: `dict`
            A dictionary for the results.

        k: `tuple`
            A key identifying the section of the field being regridded.

        srcgrid: `ESMF.Grid`
            The source grid.

        srcfield: `ESMF.Grid`
            The source field.

        srcfracfield: `ESMF.Field`
            Information about the fraction of each cell of the source
            field used in regridding.

        dstgrid: `ESMF.Grid`
            The destination grid.

        dstfield: `ESMF.Field`
            The destination field.

    :Returns:

        `None`

    """
    if not isinstance(_compute_field_mass, dict):
        raise ValueError("Expected _compute_field_mass to be a dictionary.")

    fill_value = f.fill_value(default="netCDF")

    # Calculate the mass of the source field
    srcareafield = create_Field(srcgrid, "srcareafield")
    srcmass = regrid_compute_mass_grid(
        srcfield,
        srcareafield,
        dofrac=True,
        fracfield=srcfracfield,
        uninitval=fill_value,
    )

    # Calculate the mass of the destination field
    dstareafield = create_Field(dstgrid, "dstareafield")
    dstmass = regrid_compute_mass_grid(
        dstfield, dstareafield, uninitval=fill_value
    )

    # Insert the two masses into the dictionary for comparison
    _compute_field_mass[k] = (srcmass, dstmass)

    # Release memory
    srcareafield.destroy()
    dstareafield.destroy()


def regrid_get_regridded_data(f, method, fracfield, dstfield, dstfracfield):
    """Get the regridded data.

    :Parameters:

        f: `Field`
            The source field.

        method: `str`
            The regridding method.

        fracfield: `bool`
            Whether to return the frac field or not in the case of
            conservative regridding.

        dstfield: `ESMF.Field`
            The destination field.

        dstfracfield: `ESMF.Field`
            Information about the fraction of each of the destination
            field cells involved in the regridding. For conservative
            regridding this must be taken into account.

    :Returns:

        `numpy.ndarray`
            The regridded data.

    """
    if method in conservative_regridding_methods:
        frac = dstfracfield.data.copy()
        if fracfield:
            regridded_data = frac
        else:
            frac[frac == 0.0] = 1.0
            regridded_data = np.ma.MaskedArray(
                dstfield.data / frac,
                mask=(dstfield.data == f.fill_value(default="netCDF")),
            )
    else:
        regridded_data = np.ma.MaskedArray(
            dstfield.data.copy(),
            mask=(dstfield.data == f.fill_value(default="netCDF")),
        )

    return regridded_data


def regrid_update_coordinate_references(
    f,
    dst,
    src_axis_keys,
    dst_axis_sizes,
    method,
    use_dst_mask,
    cartesian=False,
    axes=("X", "Y"),
    n_axes=2,
    src_cyclic=False,
    dst_cyclic=False,
):
    """Update the coordinate references of the regridded field.

    :Parameters:

        f: `Field`
            The regridded field. Updated in-place.

        dst: `Field` or `dict`
            The object with the destination grid for regridding.

        src_axis_keys: sequence of `str`
            The keys of the source regridding axes.

        dst_axis_sizes: sequence, optional
            The sizes of the destination axes.

        method: `bool`
            The regridding method.

        use_dst_mask: `bool`
            Whether to use the destination mask in regridding.

        cartesian: `bool`, optional
            Whether to do Cartesian regridding or spherical

        axes: sequence, optional
            Specifiers for the regridding axes.

        n_axes: `int`, optional
            The number of regridding axes.

        src_cyclic: `bool`, optional
            Whether the source longitude is cyclic for spherical
            regridding.

        dst_cyclic: `bool`, optional
            Whether the destination longitude is cyclic for spherical
            regridding.

    :Returns:

        `None`

    """
    domain_ancillaries = f.domain_ancillaries(todict=True)

    # Initialise cached value for domain_axes
    domain_axes = None

    data_axes = f.constructs.data_axes()

    for key, ref in f.coordinate_references(todict=True).items():
        ref_axes = []
        for k in ref.coordinates():
            ref_axes.extend(data_axes[k])

        if set(ref_axes).intersection(src_axis_keys):
            f.del_construct(key)
            continue

        for (
            term,
            value,
        ) in ref.coordinate_conversion.domain_ancillaries().items():
            if value not in domain_ancillaries:
                continue

            key = value

            # If this domain ancillary spans both X and Y axes
            # then regrid it, otherwise remove it
            x = f.domain_axis("X", key=True)
            y = f.domain_axis("Y", key=True)
            if (
                f.domain_ancillary(
                    filter_by_axis=(x, y),
                    axis_mode="exact",
                    key=True,
                    default=None,
                )
                is not None
            ):
                # Convert the domain ancillary into an independent
                # field
                value = f.convert(key)
                try:
                    if cartesian:
                        value.regridc(
                            dst,
                            axes=axes,
                            method=method,
                            use_dst_mask=use_dst_mask,
                            inplace=True,
                        )
                    else:
                        value.regrids(
                            dst,
                            src_cyclic=src_cyclic,
                            dst_cyclic=dst_cyclic,
                            method=method,
                            use_dst_mask=use_dst_mask,
                            inplace=True,
                        )
                except ValueError:
                    ref.coordinate_conversion.set_domain_ancillary(term, None)
                    f.del_construct(key)
                else:
                    ref.coordinate_conversion.set_domain_ancillary(term, key)
                    d_axes = data_axes[key]

                    domain_axes = f.domain_axes(
                        cached=domain_axes, todict=True
                    )

                    for k_s, new_size in zip(src_axis_keys, dst_axis_sizes):
                        domain_axes[k_s].set_size(new_size)

                    f.set_construct(
                        f._DomainAncillary(source=value),
                        key=key,
                        axes=d_axes,
                        copy=False,
                    )


def regrid_copy_coordinate_references(f, dst, dst_axis_keys):
    """Copy coordinate references from the destination field to the new,
    regridded field.

    :Parameters:

        f: `Field`
            The source field.

        dst: `Field`
            The destination field.

        dst_axis_keys: sequence of `str`
            The keys of the regridding axes in the destination field.

    :Returns:

        `None`

    """
    dst_data_axes = dst.constructs.data_axes()

    for ref in dst.coordinate_references(todict=True).values():
        axes = set()
        for key in ref.coordinates():
            axes.update(dst_data_axes[key])

        if axes and set(axes).issubset(dst_axis_keys):
            # This coordinate reference's coordinates span the X
            # and/or Y axes
            f.set_coordinate_reference(ref, parent=dst, strict=True)


def regrid_use_bounds(method):
    """Returns whether to use the bounds or not in regridding. This is
    only the case for conservative regridding.

    :Parameters:

        method: `str`
            The regridding method

    :Returns:

        `bool`

    """
    return method in conservative_regridding_methods


def regrid_update_coordinates(
    f,
    dst,
    dst_dict,
    dst_coords,
    src_axis_keys,
    dst_axis_keys,
    cartesian=False,
    dst_axis_sizes=None,
    dst_coords_2D=False,
    dst_coord_order=None,
):
    """Update the coordinates of the regridded field.

    :Parameters:

        f: `Field`
            The regridded field. Updated in-place.

        dst: Field or `dict`
            The object containing the destination grid.

        dst_dict: `bool`
            Whether dst is a dictionary.

        dst_coords: sequence
            The destination coordinates.

        src_axis_keys: sequence
            The keys of the regridding axes in the source field.

        dst_axis_keys: sequence
            The keys of the regridding axes in the destination field.

        cartesian: `bool`, optional
            Whether regridding is Cartesian of spherical, False by
            default.

        dst_axis_sizes: sequence, optional
            The sizes of the destination axes.

        dst_coords_2D: `bool`, optional
            Whether the destination coordinates are 2-d, currently only
            applies to spherical regridding.

        dst_coord_order: `list`, optional
            A list of lists specifying the ordering of the axes for
            each 2-d destination coordinate.

    :Returns:

        `None`

    """
    # NOTE: May be common ground between cartesian and spherical that
    # could save some lines of code.

    # Remove the source coordinates of new field
    for key in f.coordinates(
        filter_by_axis=src_axis_keys, axis_mode="or", todict=True
    ):
        f.del_construct(key)

    domain_axes = f.domain_axes(todict=True)

    if cartesian:
        # Insert coordinates from dst into new field
        if dst_dict:
            for k_s, coord in zip(src_axis_keys, dst_coords):
                domain_axes[k_s].set_size(coord.size)
                f.set_construct(coord, axes=[k_s])
        else:
            axis_map = {
                key_d: key_s
                for key_s, key_d in zip(src_axis_keys, dst_axis_keys)
            }

            for key_d in dst_axis_keys:
                dim = dst.dimension_coordinate(filter_by_axis=(key_d,))
                key_s = axis_map[key_d]
                domain_axes[key_s].set_size(dim.size)
                f.set_construct(dim, axes=[key_s])

            dst_data_axes = dst.constructs.data_axes()

            for aux_key, aux in dst.auxiliary_coordinates(
                filter_by_axis=dst_axis_keys,
                axis_mode="subset",
                todict=True,
            ).items():
                aux_axes = [
                    axis_map[key_d] for key_d in dst_data_axes[aux_key]
                ]
                f.set_construct(aux, axes=aux_axes)
    else:
        # Give destination grid latitude and longitude standard names
        dst_coords[0].standard_name = "longitude"
        dst_coords[1].standard_name = "latitude"

        # Insert 'X' and 'Y' coordinates from dst into new field
        for axis_key, axis_size in zip(src_axis_keys, dst_axis_sizes):
            domain_axes[axis_key].set_size(axis_size)

        if dst_dict:
            if dst_coords_2D:
                for coord, coord_order in zip(dst_coords, dst_coord_order):
                    axis_keys = [src_axis_keys[index] for index in coord_order]
                    f.set_construct(coord, axes=axis_keys)
            else:
                for coord, axis_key in zip(dst_coords, src_axis_keys):
                    f.set_construct(coord, axes=[axis_key])

        else:
            for src_axis_key, dst_axis_key in zip(
                src_axis_keys, dst_axis_keys
            ):
                dim_coord = dst.dimension_coordinate(
                    filter_by_axis=(dst_axis_key,), default=None
                )
                if dim_coord is not None:
                    f.set_construct(dim_coord, axes=[src_axis_key])

                for aux in dst.auxiliary_coordinates(
                    filter_by_axis=(dst_axis_key,),
                    axis_mode="exact",
                    todict=True,
                ).values():
                    f.set_construct(aux, axes=[src_axis_key])

            for aux_key, aux in dst.auxiliary_coordinates(
                filter_by_axis=dst_axis_keys,
                axis_mode="subset",
                todict=True,
            ).items():
                aux_axes = dst.get_data_axes(aux_key)
                if aux_axes == tuple(dst_axis_keys):
                    f.set_construct(aux, axes=src_axis_keys)
                else:
                    f.set_construct(aux, axes=src_axis_keys[::-1])

    # Copy names of dimensions from destination to source field
    if not dst_dict:
        dst_domain_axes = dst.domain_axes(todict=True)
        for src_axis_key, dst_axis_key in zip(src_axis_keys, dst_axis_keys):
            ncdim = dst_domain_axes[dst_axis_key].nc_get_dimension(None)
            if ncdim is not None:
                domain_axes[src_axis_key].nc_set_dimension(ncdim)


def grids_have_same_masks(grid0, grid1):
    """Whether two `ESMF.Grid` instances have identical masks.

    :Parameters:

        grid0, grid1: `ESMF.Grid`, `ESMF.Grid`
            The `ESMF` Grid instances to be compared

    :Returns:

        `bool`
            Whether or not the Grids have identical masks.

    """
    mask0 = grid0.mask
    mask1 = grid1.mask

    if len(mask0) != len(mask1):
        return False

    for a, b in zip(mask0, mask1):
        if not np.array_equal(a, b):
            return False

    return True


def grids_have_same_coords(grid0, grid1):
    """Whether two `ESMF.Grid` instances have identical coordinates.

    :Parameters:

        grid0, grid1: `ESMF.Grid`, `ESMF.Grid`
            The `ESMF` Grid instances to be compared

    :Returns:

        `bool`
            Whether or not the Grids have identical coordinates.

    """
    coords0 = grid0.coords
    coords1 = grid1.coords

    if len(coords0) != len(coords1):
        return False

    for c, d in zip(coords0, coords1):
        if len(c) != len(d):
            return False

        for a, b in zip(c, d):
            if not np.array_equal(a, b):
                return False

    return True


def regrid_initialize():
    """Initialize `ESMF`.

    Initialise the `ESMF` manager. Whether logging is enabled or not
    is determined by cf.regrid_logging. If it is then logging takes
    place after every call to `ESMF`.

    :Returns:

        `ESMF.Manager`
            A singleton instance of the `ESMF` manager.

    """
    if not _found_ESMF:
        raise RuntimeError(
            "Regridding methods will not work unless "
            "the ESMF library is installed"
        )

    return ESMF.Manager(debug=bool(regrid_logging()))


def create_Regrid(
    srcfield,
    dstfield,
    srcfracfield,
    dstfracfield,
    method,
    ignore_degenerate,
):
    """Create an `ESMF` regrid operator.

    :Parameters:

        srcfield: `ESMF.Field`
            The source Field with an associated grid to be used
            for regridding.

        dstfield: `ESMF.Field`
            The destination Field with an associated grid to be
            used for regridding.

        srcfracfield: `ESMF.Field`
            A Field to hold the fraction of the source field that
            contributes to conservative regridding.

        dstfracfield: `ESMF.Field`
            A Field to hold the fraction of the source field that
            contributes to conservative regridding.

        method: `str`
            Setting as follows gives a corresponding result of:

            ===================  =================================
            *method*             Description
            ===================  =================================
            'conservative_1st'   Use first-order conservative
                                 regridding.

            'conservative'       Alias for 'conservative_1st'.

            'conservative_2nd'   Use second-order conservative
                                 regridding.

            'linear'             Use (multi)linear interpolation.

            'bilinear'           Deprecated alias for 'linear'.

            'patch'              Use higher-order patch recovery.

            'nearest_stod'       Use nearest source to
                                 destination interpolation.

            'nearest_dtos'       Use nearest destination to
                                 source interpolation.
            ===================  =================================

        ignore_degenerate: `bool`
            Whether to check for degenerate points.

    :Returns:

        `ESMF.Regrid`
            The `ESMF` regrid operator.

    """
    regrid_method = regrid_method_map.get(method)
    if regrid_method is None:
        raise ValueError(f"Regrid method {method!r} not recognised.")

    # Initialise the regridder. This also creates the weights
    # needed for the regridding.
    return ESMF.Regrid(
        srcfield,
        dstfield,
        regrid_method=regrid_method,
        src_mask_values=np.array([0], dtype="int32"),
        dst_mask_values=np.array([0], dtype="int32"),
        src_frac_field=srcfracfield,
        dst_frac_field=dstfracfield,
        unmapped_action=ESMF.UnmappedAction.IGNORE,
        ignore_degenerate=ignore_degenerate,
    )


def create_Grid(
    coords,
    use_bounds,
    mask=None,
    cartesian=False,
    cyclic=False,
    coords_2D=False,
    coord_order=None,
):
    """Create an `ESMF` Grid.

    :Parameters:

        coords: sequence
            The coordinates if not Cartesian it is assume that the
            first is longitude and the second is latitude.

        use_bounds: `bool`
            Whether to populate the grid corners with information from
            the bounds or not.

        mask: `numpy.ndarray`, optional
            An optional numpy array of booleans containing the grid
            points to mask. Where the elements of mask are True the
            output grid is masked.

        cartesian: `bool`, optional
            Whether to create a Cartesian grid or a spherical one,
            False by default.

        cyclic: `bool`, optional
            Whether or not the longitude (if present) is cyclic. If
            None the a check for cyclicity is made from the
            bounds. None by default.

        coords_2D: `bool`, optional
            Whether the coordinates are 2-d or not. Presently only
            works for spherical coordinates. False by default.

        coord_order: sequence, optional
            Two tuples one indicating the order of the x and y axes
            for 2-d longitude, one for 2-d latitude.

    :Returns:

        `ESMF.Grid`
            The `ESMF` grid.

    """
    if not cartesian:
        lon = coords[0]
        lat = coords[1]
        if not coords_2D:
            # Get the shape of the grid
            shape = [lon.size, lat.size]
        else:
            x_order = coord_order[0]
            y_order = coord_order[1]
            # Get the shape of the grid
            shape = lon.transpose(x_order).shape
            if lat.transpose(y_order).shape != shape:
                raise ValueError(
                    "The longitude and latitude coordinates"
                    " must have the same shape."
                )

        if use_bounds:
            if not coords_2D:
                # Get the bounds
                x_bounds = lon.get_bounds()
                y_bounds = lat.get_bounds().clip(-90, 90, "degrees").array

                # If cyclic not set already, check for cyclicity
                if cyclic is None:
                    cyclic = abs(
                        x_bounds.datum(-1) - x_bounds.datum(0)
                    ) == Data(360, "degrees")

                x_bounds = x_bounds.array
            else:
                # Get the bounds
                x_bounds = lon.get_bounds()
                y_bounds = lat.get_bounds().clip(-90, 90, "degrees")
                n = x_bounds.shape[0]
                m = x_bounds.shape[1]
                x_bounds = x_bounds.array
                y_bounds = y_bounds.array

                tmp_x = np.empty((n + 1, m + 1))
                tmp_x[:n, :m] = x_bounds[:, :, 0]
                tmp_x[:n, m] = x_bounds[:, -1, 1]
                tmp_x[n, :m] = x_bounds[-1, :, 3]
                tmp_x[n, m] = x_bounds[-1, -1, 2]

                tmp_y = np.empty((n + 1, m + 1))
                tmp_y[:n, :m] = y_bounds[:, :, 0]
                tmp_y[:n, m] = y_bounds[:, -1, 1]
                tmp_y[n, :m] = y_bounds[-1, :, 3]
                tmp_y[n, m] = y_bounds[-1, -1, 2]

                x_bounds = tmp_x
                y_bounds = tmp_y

        else:
            if not coords_2D:
                # If cyclicity not set already, check for cyclicity
                if cyclic is None:
                    try:
                        x_bounds = lon.get_bounds()
                        cyclic = abs(
                            x_bounds.datum(-1) - x_bounds.datum(0)
                        ) == Data(360, "degrees")
                    except ValueError:
                        pass

        # Create empty grid
        max_index = np.array(shape, dtype="int32")
        if use_bounds:
            staggerLocs = [ESMF.StaggerLoc.CORNER, ESMF.StaggerLoc.CENTER]
        else:
            staggerLocs = [ESMF.StaggerLoc.CENTER]

        if cyclic:
            grid = ESMF.Grid(
                max_index, num_peri_dims=1, staggerloc=staggerLocs
            )
        else:
            grid = ESMF.Grid(max_index, staggerloc=staggerLocs)

        # Populate grid centres
        x, y = 0, 1
        gridXCentre = grid.get_coords(x, staggerloc=ESMF.StaggerLoc.CENTER)
        gridYCentre = grid.get_coords(y, staggerloc=ESMF.StaggerLoc.CENTER)
        if not coords_2D:
            gridXCentre[...] = lon.array.reshape((lon.size, 1))
            gridYCentre[...] = lat.array.reshape((1, lat.size))
        else:
            gridXCentre[...] = lon.transpose(x_order).array
            gridYCentre[...] = lat.transpose(y_order).array

        # Populate grid corners if there are bounds
        if use_bounds:
            gridCorner = grid.coords[ESMF.StaggerLoc.CORNER]
            if not coords_2D:
                if cyclic:
                    gridCorner[x][...] = x_bounds[:, 0].reshape(lon.size, 1)
                else:
                    n = x_bounds.shape[0]
                    tmp_x = np.empty(n + 1)
                    tmp_x[:n] = x_bounds[:, 0]
                    tmp_x[n] = x_bounds[-1, 1]
                    gridCorner[x][...] = tmp_x.reshape(lon.size + 1, 1)

                n = y_bounds.shape[0]
                tmp_y = np.empty(n + 1)
                tmp_y[:n] = y_bounds[:, 0]
                tmp_y[n] = y_bounds[-1, 1]
                gridCorner[y][...] = tmp_y.reshape(1, lat.size + 1)
            else:
                gridCorner = grid.coords[ESMF.StaggerLoc.CORNER]
                x_bounds = x_bounds.transpose(x_order)
                y_bounds = y_bounds.transpose(y_order)
                if cyclic:
                    x_bounds = x_bounds[:-1, :]
                    y_bounds = y_bounds[:-1, :]

                gridCorner[x][...] = x_bounds
                gridCorner[y][...] = y_bounds
    else:
        # Test the dimensionality of the list of coordinates
        ndim = len(coords)
        if ndim < 1 or ndim > 3:
            raise ValueError(
                "Cartesian grid must have between 1 and 3 dimensions."
            )

        # For 1-d conservative regridding add an extra dimension of
        # size 1
        if ndim == 1:
            if not use_bounds:
                # For 1-d non-conservative regridding, the extra
                # dimension should already have been added in the
                # calling function.
                raise ValueError(
                    "Cannot create a Cartesian grid from "
                    "one dimension coordinate with no bounds."
                )
            coords = [
                DimensionCoordinate(
                    data=Data(0),
                    bounds=Data(
                        [
                            np.finfo("float32").epsneg,
                            np.finfo("float32").eps,
                        ]
                    ),
                )
            ] + coords
            if mask is not None:
                mask = mask[None, :]
            ndim = 2

        shape = list()
        for coord in coords:
            shape.append(coord.size)

        # Initialise the grid
        max_index = np.array(shape, dtype="int32")
        if use_bounds:
            if ndim < 3:
                staggerLocs = [
                    ESMF.StaggerLoc.CORNER,
                    ESMF.StaggerLoc.CENTER,
                ]
            else:
                staggerLocs = [
                    ESMF.StaggerLoc.CENTER_VCENTER,
                    ESMF.StaggerLoc.CORNER_VFACE,
                ]
        else:
            if ndim < 3:
                staggerLocs = [ESMF.StaggerLoc.CENTER]
            else:
                staggerLocs = [ESMF.StaggerLoc.CENTER_VCENTER]

        grid = ESMF.Grid(
            max_index, coord_sys=ESMF.CoordSys.CART, staggerloc=staggerLocs
        )

        # Populate the grid centres
        for d in range(0, ndim):
            if ndim < 3:
                gridCentre = grid.get_coords(
                    d, staggerloc=ESMF.StaggerLoc.CENTER
                )
            else:
                gridCentre = grid.get_coords(
                    d, staggerloc=ESMF.StaggerLoc.CENTER_VCENTER
                )
            gridCentre[...] = coords[d].array.reshape(
                [shape[d] if x == d else 1 for x in range(0, ndim)]
            )

        # Populate grid corners
        if use_bounds:
            if ndim < 3:
                gridCorner = grid.coords[ESMF.StaggerLoc.CORNER]
            else:
                gridCorner = grid.coords[ESMF.StaggerLoc.CORNER_VFACE]

            for d in range(0, ndim):
                # boundsD = coords[d].get_bounds(create=True).array
                boundsD = coords[d].get_bounds(None)
                if boundsD is None:
                    boundsD = coords[d].create_bounds()

                boundsD = boundsD.array

                if shape[d] > 1:
                    tmp = np.empty(shape[d] + 1)
                    tmp[0:-1] = boundsD[:, 0]
                    tmp[-1] = boundsD[-1, 1]
                    boundsD = tmp

                gridCorner[d][...] = boundsD.reshape(
                    [shape[d] + 1 if x == d else 1 for x in range(0, ndim)]
                )

    # Add the mask if appropriate
    if mask is not None:
        gmask = grid.add_item(ESMF.GridItem.MASK)
        gmask[...] = 1
        gmask[mask] = 0

    return grid


def create_Field(grid, name):
    """Create an `ESMF` Field.

    :Parameters:

        grid: `ESMF.Grid`
            The `ESMF` grid to use in creating the field.

        name: `str`
            The name to give the `ESMF` Field.

    :Returns:

        `ESMF.Field`
            The `ESMF` Field for use as a source or destination Field
            in regridding.

    """
    return ESMF.Field(grid, name)


def run_Regrid(regrid, srcfield, dstfield):
    """Call an `ESMF.Regrid` instance to perform regridding.

    :Parameters:

        regrid: `ESMF.Regrid`
            The `ESMF` regrid operator

        srcfield: `ESMF.Field`
            The source Field with an associated grid to be used for
            regridding.

        dstfield: `ESMF.Field`
            The destination Field with an associated grid to be used
            for regridding.

    :Returns:

        `ESMF.Field`
            The regridded Field

    """
    return regrid(srcfield, dstfield, zero_region=ESMF.Region.SELECT)


def regrid_create_operator(regrid, name, parameters):
    """Create a new `RegridOperator` instance.

    :Parameters:

        regrid: `ESMF.Regrid`
            The `ESMF` regridding operator between two fields.

        name: `str`
            A descriptive name for the operator.

        parameters: `dict`
            Parameters that describe the complete coordinate system of
            the destination grid.

    :Returns:

        `RegridOperator`
            The new regrid operator.

    """
    return RegridOperator(regrid, name, **parameters)


def regrid_get_operator_method(operator, method):
    """Return the regridding method of a regridding operator.

    :Parameters:

        operator: `RegridOperator`
            The regridding operator.

        method: `str` or `None`
            A regridding method. If `None` then ignored. If a `str`
            then an exception is raised if it not equivalent to the
            regridding operator's method.

    :Returns:

        `string`
            The operator's regridding method.

    """
    if method is None:
        method = operator.method
    elif not operator.check_method(method):
        raise ValueError(
            f"Method {method!r} does not match the method of the "
            f"regridding operator: {operator.method!r}"
        )

    return method



import cfdm


class GatheredArray(cfdm.GatheredArray):
    """An underlying gathered array.

    Compression by gathering combines axes of a multidimensional array
    into a new, discrete axis whilst omitting the missing values and
    thus reducing the number of values that need to be stored.

    The information needed to uncompress the data is stored in a "list
    variable" that gives the indices of the required points.

    .. versionadded:: 3.0.0

    """


# --- End: class



import cfdm
from numpy import empty as numpy_empty
from numpy import result_type as numpy_result_type

from . import Bounds, mixin
from .data.data import Data
from .decorators import (
    _deprecated_kwarg_check,
    _inplace_enabled,
    _inplace_enabled_define_and_cleanup,
)
from .functions import _DEPRECATION_ERROR_ATTRIBUTE, _DEPRECATION_ERROR_KWARGS
from .timeduration import TimeDuration
from .units import Units


class DimensionCoordinate(
    mixin.Coordinate, mixin.PropertiesDataBounds, cfdm.DimensionCoordinate
):
    """A dimension coordinate construct of the CF data model.

    A dimension coordinate construct provides information which locate
    the cells of the domain and which depend on a subset of the domain
    axis constructs. The dimension coordinate construct is able to
    unambiguously describe cell locations because a domain axis can be
    associated with at most one dimension coordinate construct, whose
    data array values must all be non-missing and strictly
    monotonically increasing or decreasing. They must also all be of
    the same numeric data type. If cell bounds are provided, then each
    cell must have exactly two vertices. CF-netCDF coordinate
    variables and numeric scalar coordinate variables correspond to
    dimension coordinate constructs.

    The dimension coordinate construct consists of a data array of the
    coordinate values which spans a subset of the domain axis
    constructs, an optional array of cell bounds recording the extents
    of each cell (stored in a `Bounds` object), and properties to
    describe the coordinates. An array of cell bounds spans the same
    domain axes as its coordinate array, with the addition of an extra
    dimension whose size is that of the number of vertices of each
    cell. This extra dimension does not correspond to a domain axis
    construct since it does not relate to an independent axis of the
    domain. Note that, for climatological time axes, the bounds are
    interpreted in a special way indicated by the cell method
    constructs.

    **NetCDF interface**

    {{netcdf variable}}

    """

    def __new__(cls, *args, **kwargs):
        """Store component classes."""
        instance = super().__new__(cls)
        instance._Bounds = Bounds
        return instance

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)

    def _centre(self, period):
        """It assumed, but not checked, that the period has been set.

        .. seealso:: `roll`

        """
        if self.direction():
            mx = self.data[-1]
        else:
            mx = self.data[0]

        return ((mx // period) * period).squeeze()

    def _infer_direction(self):
        """Return True if a coordinate is increasing, otherwise return
        False.

        A dimension coordinate construct is considered to be increasing if
        its data array values are increasing in index space, or if it has
        no data nor bounds.

        If the direction can not be inferred from the data not bounds then
        the coordinate's units are used.

        The direction is inferred from the coordinate's data array values
        or its from coordinates. It is not taken directly from its
        `cf.Data` object.

        :Returns:

            `bool`
                Whether or not the coordinate is increasing.

        **Examples**

        >>> c.array
        array([  0  30  60])
        >>> c._get_direction()
        True
        >>> c.array
        array([15])
        >>> c.bounds.array
        array([  30  0])
        >>> c._get_direction()
        False

        """
        data = self.get_data(None, _fill_value=False)
        if data is not None:
            # Infer the direction from the data
            if data._size > 1:
                data = data[0:2].array
                return bool(
                    data.item(
                        0,
                    )
                    < data.item(
                        1,
                    )
                )

        # Still here?
        data = self.get_bounds_data(None, _fill_value=False)
        if data is not None:
            # Infer the direction from the bounds
            b = data[(0,) * (data.ndim - 1)].array
            return bool(
                b.item(
                    0,
                )
                < b.item(
                    1,
                )
            )

        # Still here? Then infer the direction from the units.
        return not self.Units.ispressure

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    #    @property
    #    def cellsize(self):
    #        '''The cell sizes.
    #
    #    :Returns:
    #
    #        `Data`
    #            The size for each cell.
    #
    #    **Examples**
    #
    #    >>> print(c.bounds)
    #    <CF Bounds: latitude(3, 2) degrees_north>
    #    >>> print(c.bounds.array)
    #    [[-90. -87.]
    #     [-87. -80.]
    #     [-80. -67.]]
    #    >>> print(d.cellsize)
    #    <CF Data(3): [3.0, 7.0, 13.0] degrees_north>
    #    >>> print(d.cellsize.array)
    #    [  3.   7.  13.]
    #    >>> print(c.sin().cellsize.array)
    #    [ 0.00137047  0.01382178  0.0643029 ]
    #
    #    >>> del(c.bounds)
    #    >>> c.cellsize
    #    AttributeError: Can't get cell sizes when coordinates have no bounds
    #
    #        '''
    #        cells = self.get_bounds_data(None)
    #        if cells is None:
    #            raise AttributeError(
    #                "Can't get cell sizes when coordinates have no bounds")
    #
    #        if self.direction():
    #            cells = cells[:, 1] - cells[:, 0]
    #        else:
    #            cells = cells[:, 0] - cells[:, 1]
    #
    #        cells.squeeze(1, inplace=True)
    #
    #        return cells

    @property
    def decreasing(self):
        """True if the dimension coordinate is decreasing, otherwise
        False.

        A dimension coordinate is increasing if its coordinate values are
        increasing in index space.

        The direction is inferred from one of, in order of precedence:

        * The data array
        * The bounds data array
        * The `units` CF property

        :Returns:

            `bool`
                Whether or not the coordinate is decreasing.

        **Examples**

        >>> c.decreasing
        False
        >>> c.flip().increasing
        True

        """
        return not self.direction()

    @property
    def increasing(self):
        """`True` for dimension coordinate constructs, `False`
        otherwise.

        A dimension coordinate is increasing if its coordinate values are
        increasing in index space.

        The direction is inferred from one of, in order of precedence:

        * The data array
        * The bounds data array
        * The `units` CF property

        :Returns:

            `bool`
                Whether or not the coordinate is increasing.

        **Examples**

        >>> c.decreasing
        False
        >>> c.flip().increasing
        True

        """
        return self.direction()

    #    @property
    #    def lower_bounds(self):
    #        '''The lower dimension coordinate bounds in a `cf.Data` object.
    #
    #    .. seealso:: `bounds`, `upper_bounds`
    #
    #    **Examples**
    #
    #    >>> print(c.bounds.array)
    #    [[ 5  3]
    #     [ 3  1]
    #     [ 1 -1]]
    #    >>> c.lower_bounds
    #    <CF Data(3): [3, 1, -1]>
    #    >>> print(c.lower_bounds.array)
    #    [ 3  1 -1]
    #
    #        '''
    #        data = self.get_bounds_data(None)
    #        if data is None:
    #            raise ValueError(
    #                "Can't get lower bounds when there are no bounds")
    #
    #        if self.direction():
    #            i = 0
    #        else:
    #            i = 1
    #
    #        out = data[..., i]
    #        out.squeeze(1, inplace=True)
    #        return out

    #
    #    @property
    #    def upper_bounds(self):
    #        '''The upper dimension coordinate bounds in a `cf.Data` object.
    #
    #    .. seealso:: `bounds`, `lower_bounds`
    #
    #    **Examples**
    #
    #    >>> print(c.bounds.array)
    #    [[ 5  3]
    #     [ 3  1]
    #     [ 1 -1]]
    #    >>> c.upper_bounds
    #    <CF Data(3): [5, 3, 1]>
    #    >>> print(c.upper_bounds.array)
    #    [5  3  1]
    #
    #        '''
    #        data = self.get_bounds_data(None)
    #        if data is None:
    #            raise ValueError(
    #                "Can't get upper bounds when there are no bounds")
    #
    #        if self.direction():
    #            i = 1
    #        else:
    #            i = 0
    #
    #        return data[..., i].squeeze(1)

    def direction(self):
        """Return True if the dimension coordinate values are
        increasing, otherwise return False.

        Dimension coordinates values are increasing if its coordinate
        values are increasing in index space.

        The direction is inferred from one of, in order of precedence:

        * The data array
        * The bounds data array
        * The `units` CF property

        :Returns:

            `bool`
                Whether or not the coordinate is increasing.

        **Examples**

        >>> c.array
        array([  0  30  60])
        >>> c.direction()
        True

        >>> c.bounds.array
        array([  30  0])
        >>> c.direction()
        False

        """
        _direction = self._custom.get("direction")
        if _direction is not None:
            return _direction

        _direction = self._infer_direction()
        self._custom["direction"] = _direction

        return _direction

    def create_bounds(
        self, bound=None, cellsize=None, flt=0.5, max=None, min=None
    ):
        """Create cell bounds.

        :Parameters:

            bound: optional
                If set to a value larger (smaller) than the largest
                (smallest) coordinate value then bounds are created which
                include this value and for which each coordinate is in the
                centre of its bounds.

            cellsize: optional
                Define the exact size of each cell that is
                created. Created cells are allowed to overlap do not have
                to be contigious. The *cellsize* parameter may be one of:

                  * A data-like scalar (see below) that defines the cell size,
                    either in the same units as the coordinates or in the
                    units provided. Note that in this case, the position of
                    each coordinate within the its cell is controlled by the
                    *flt* parameter.

                    *Parameter example:*
                        To specify cellsizes of 10, in the same units as the
                        coordinates: ``cellsize=10``.

                    *Parameter example:*
                        To specify cellsizes of 1 day: ``cellsize=cf.Data(1,
                        'day')`` (see `cf.Data` for details).

                    *Parameter example:*
                        For coordinates ``1, 2, 10``, setting ``cellsize=1``
                        will result in bounds of ``(0.5, 1.5), (1.5, 2.5),
                        (9.5, 10.5)``.

                    *Parameter example:*
                        For coordinates ``1, 2, 10`` kilometres, setting
                        ``cellsize=cf.Data(5000, 'm')`` will result in bounds
                        of ``(-1.5, 3.5), (-0.5, 4.5), (7.5, 12.5)`` (see
                        `cf.Data` for details).

                    *Parameter example:*
                        For decreasing coordinates ``2, 0, -12`` setting,
                        ``cellsize=2`` will result in bounds of ``(3, 1),
                        (1, -1), (-11, -13)``.

                  * A `cf.TimeDuration` defining the cell size. Only
                    applicable to reference time coordinates. It is possible
                    to "anchor" the cell bounds via the `cf.TimeDuration`
                    parameters. For example, to specify cell size of one
                    calendar month, starting and ending on the 15th day:
                    ``cellsize=cf.M(day=15)`` (see `cf.M` for details). Note
                    that the *flt* parameter is ignored in this case.

                    *Parameter example:*
                        For coordinates ``1984-12-01 12:00, 1984-12-02
                        12:00, 2000-04-15 12:00`` setting,
                        ``cellsize=cf.D()`` will result in bounds of
                        ``(1984-12-01, 1984-12-02), (1984-12-02,
                        1984-12-03), (2000-05-15, 2000-04-16)`` (see `cf.D`
                        for details).

                    *Parameter example:*
                        For coordinates ``1984-12-01, 1984-12-02,
                        2000-04-15`` setting, ``cellsize=cf.D()`` will
                        result in bounds of ``(1984-12-01, 1984-12-02),
                        (1984-12-02, 1984-12-03), (2000-05-15, 2000-04-16)``
                        (see `cf.D` for details).

                    *Parameter example:*
                        For coordinates ``1984-12-01, 1984-12-02,
                        2000-04-15`` setting, ``cellsize=cf.D(hour=12)``
                        will result in bounds of ``(1984-11:30 12:00,
                        1984-12-01 12:00), (1984-12-01 12:00, 1984-12-02
                        12:00), (2000-05-14 12:00, 2000-04-15 12:00)`` (see
                        `cf.D` for details).

                    *Parameter example:*
                        For coordinates ``1984-12-16 12:00, 1985-01-16
                        12:00`` setting, ``cellsize=cf.M()`` will result in
                        bounds of ``(1984-12-01, 1985-01-01), (1985-01-01,
                        1985-02-01)`` (see `cf.M` for details).

                    *Parameter example:*
                        For coordinates ``1984-12-01 12:00, 1985-01-01
                        12:00`` setting, ``cellsize=cf.M()`` will result in
                        bounds of ``(1984-12-01, 1985-01-01), (1985-01-01,
                        1985-02-01)`` (see `cf.M` for details).

                    *Parameter example:*
                        For coordinates ``1984-12-01 12:00, 1985-01-01
                        12:00`` setting, ``cellsize=cf.M(day=20)`` will
                        result in bounds of ``(1984-11-20, 1984-12-20),
                        (1984-12-20, 1985-01-20)`` (see `cf.M` for details).

                    *Parameter example:*
                        For coordinates ``1984-03-01, 1984-06-01`` setting,
                        ``cellsize=cf.Y()`` will result in bounds of
                        ``(1984-01-01, 1985-01-01), (1984-01-01,
                        1985-01-01)`` (see `cf.Y` for details). Note that in
                        this case each cell has the same bounds. This
                        because ``cf.Y()`` is equivalent to ``cf.Y(month=1,
                        day=1)`` and the closest 1st January to both
                        coordinates is 1st January 1984.

                {+data-like-scalar} TODO

            flt: `float`, optional
                When creating cells with sizes specified by the *cellsize*
                parameter, define the fraction of the each cell which is
                less its coordinate value. By default *flt* is 0.5, so that
                each cell has its coordinate at it's centre. Ignored if
                *cellsize* is not set.

                *Parameter example:*
                    For coordinates ``1, 2, 10``, setting ``cellsize=1,
                    flt=0.5`` will result in bounds of ``(0.5, 1.5), (1.5,
                    2.5), (9.5, 10.5)``.

                *Parameter example:*
                    For coordinates ``1, 2, 10``, setting ``cellsize=1,
                    flt=0.25`` will result in bounds of ``(0.75, 1.75),
                    (1.75, 2.75), (9.75, 10.75)``.

                *Parameter example:*
                    For decreasing coordinates ``2, 0, -12``, setting
                    ``cellsize=6, flt=0.9`` will result in bounds of
                    ``(2.6, -3.4), (0.6, -5.4), (-11.4, -17.4)``.

            max: optional
                Limit the created bounds to be no more than this number.

                *Parameter example:*
                    To ensure that all latitude bounds are at most 90:
                    ``max=90``.

            min: optional
                Limit the created bounds to be no less than this number.

                *Parameter example:*
                    To ensure that all latitude bounds are at least -90:
                    ``min=-90``.

        :Returns:

            `Bounds`
                The newly-created coordinate cell bounds object.

        **Examples**

        >>> c.create_bounds()
        >>> c.create_bounds(bound=-9000.0)

        """
        array = self.array
        size = array.size

        if cellsize is not None:
            if bound:
                raise ValueError(
                    "bound parameter can't be True when setting the "
                    "cellsize parameter"
                )

            if not isinstance(cellsize, TimeDuration):
                # ----------------------------------------------------
                # Create bounds based on cell sizes defined by a
                # data-like object
                #
                # E.g. cellsize=10
                #      cellsize=cf.Data(1, 'day')
                # ----------------------------------------------------
                cellsize = Data.asdata(abs(cellsize))
                if cellsize.Units:
                    err_msg = (
                        "Can't create bounds because the bound units "
                        f"({cellsize.Units}) are not compatible with "
                        f"the coordinate units ({self.Units})."
                    )
                    if self.Units.isreftime:
                        if not cellsize.Units.istime:
                            raise ValueError(err_msg)
                        cellsize.Units = Units(self.Units._utime.units)
                    else:
                        if not cellsize.Units.equivalent(self.Units):
                            raise ValueError(err_msg)
                        cellsize.Units = self.Units
                cellsize = cellsize.datum()

                cellsize0 = cellsize * flt
                cellsize1 = cellsize * (1 - flt)
                if not self.direction():
                    cellsize0, cellsize1 = -cellsize1, -cellsize0

                bounds = numpy_empty((size, 2), dtype=array.dtype)
                bounds[:, 0] = array - cellsize0
                bounds[:, 1] = array + cellsize1
            else:
                # ----------------------------------------------------
                # Create bounds based on cell sizes defined by a
                # TimeDuration object
                #
                # E.g. cellsize=cf.s()
                #      cellsize=cf.m()
                #      cellsize=cf.h()
                #      cellsize=cf.D()
                #      cellsize=cf.M()
                #      cellsize=cf.Y()
                #      cellsize=cf.D(hour=12)
                #      cellsize=cf.M(day=16)
                #      cellsize=cf.M(2)
                #      cellsize=cf.M(2, day=15, hour=12)
                # ----------------------------------------------------
                if not self.Units.isreftime:
                    raise ValueError(
                        "Can't create reference time bounds for "
                        f"non-reference time coordinates: {self.Units!r}"
                    )

                bounds = numpy_empty((size, 2), dtype=object)

                cellsize_bounds = cellsize.bounds
                direction = bool(self.direction())

                for c, b in zip(self.datetime_array, bounds):
                    b[...] = cellsize_bounds(c, direction=direction)
        else:
            if bound is None:
                # ----------------------------------------------------
                # Creat Voronoi bounds
                # ----------------------------------------------------
                if size < 2:
                    raise ValueError(
                        "Can't create bounds for Voronoi cells from one value"
                    )

                bounds_1d = [
                    array.item(
                        0,
                    )
                    * 1.5
                    - array.item(
                        1,
                    )
                    * 0.5
                ]
                bounds_1d.extend((array[0:-1] + array[1:]) * 0.5)
                bounds_1d.append(
                    array.item(
                        -1,
                    )
                    * 1.5
                    - array.item(
                        -2,
                    )
                    * 0.5
                )

                dtype = type(bounds_1d[0])

                if max is not None:
                    if self.direction():
                        bounds_1d[-1] = max
                    else:
                        bounds_1d[0] = max
                if min is not None:
                    if self.direction():
                        bounds_1d[0] = min
                    else:
                        bounds_1d[-1] = min

            else:
                # ----------------------------------------------------
                # Create
                # ----------------------------------------------------
                direction = self.direction()
                if not direction and size > 1:
                    array = array[::-1]

                bounds_1d = [bound]
                if bound <= array.item(
                    0,
                ):
                    for i in range(size):
                        bound = (
                            2.0
                            * array.item(
                                i,
                            )
                            - bound
                        )
                        bounds_1d.append(bound)
                elif bound >= array.item(
                    -1,
                ):
                    for i in range(size - 1, -1, -1):
                        bound = (
                            2.0
                            * array.item(
                                i,
                            )
                            - bound
                        )
                        bounds_1d.append(bound)

                    bounds_1d = bounds_1d[::-1]
                else:
                    raise ValueError("bad bound value")

                dtype = type(bounds_1d[-1])

                if not direction:
                    bounds_1d = bounds_1d[::-1]

            bounds = numpy_empty((size, 2), dtype=dtype)
            bounds[:, 0] = bounds_1d[:-1]
            bounds[:, 1] = bounds_1d[1:]

        # Create coordinate bounds object
        bounds = Bounds(data=Data(bounds, units=self.Units), copy=False)

        return bounds

    @_deprecated_kwarg_check("i")
    @_inplace_enabled(default=False)
    def flip(self, axes=None, inplace=False, i=False):
        """Flips the dimension coordinate, that is reverses its
        direction."""
        d = _inplace_enabled_define_and_cleanup(self)

        super(DimensionCoordinate, d).flip(axes=axes, inplace=True)

        direction = d._custom.get("direction")
        if direction is not None:
            d._custom["direction"] = not direction

        return d

    def get_bounds(self, default=ValueError(), **kwargs):
        """Return the bounds.

        .. versionadded:: 3.0.0

        .. seealso:: `bounds`, `create_bounds', `get_data`,
                     `del_bounds`, `has_bounds`, `set_bounds`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if bounds
                have not been set.

                {{default Exception}}

        :Returns:

            `Bounds`
                The bounds.

        **Examples**

        >>> b = Bounds(data=cfdm.Data(range(10).reshape(5, 2)))
        >>> c.set_bounds(b)
        >>> c.has_bounds()
        True
        >>> c.get_bounds()
        <Bounds: (5, 2) >
        >>> b = c.del_bounds()
        >>> b
        <Bounds: (5, 2) >
        >>> c.has_bounds()
        False
        >>> print(c.get_bounds(None))
        None
        >>> print(c.del_bounds(None))
        None

        """
        if kwargs:
            _DEPRECATION_ERROR_KWARGS(
                self,
                "get_bounds",
                kwargs,
                "Bounds creation now uses the 'create_bounds' and "
                "'set_bounds' methods.",
            )  # pragma: no cover

        return super().get_bounds(default=default)

    #    def autoperiod(self, verbose=False):
    #        '''TODO Set dimensions to be cyclic.
    #
    #    TODO A dimension is set to be cyclic if it has a unique longitude (or
    #    grid longitude) dimension coordinate construct with bounds and the
    #    first and last bounds values differ by 360 degrees (or an
    #    equivalent amount in other units).
    #
    #    .. versionadded:: 3.0.0
    #
    #    .. seealso:: `isperiodic`, `period`
    #
    #    :Parameters:
    #
    #        TODO
    #
    #    :Returns:
    #
    #       `bool`
    #
    #    **Examples**
    #
    #    >>> f.autocyclic()
    #
    #        '''
    #        if not self.Units.islongitude:
    #            if verbose:
    #                print(0)
    #            if (self.get_property('standard_name', None) not in
    #                    ('longitude', 'grid_longitude')):
    #                if verbose:
    #                    print(1)
    #                return False
    #
    #        bounds = self.get_bounds(None)
    #        if bounds is None:
    #            if verbose:
    #                print(2)
    #            return False
    #
    #        bounds_data = bounds.get_data(None)
    #        if bounds_data is None:
    #            if verbose:
    #                print(3)
    #            return False
    #
    #        bounds = bounds_data.array
    #
    #        period = Data(360.0, units='degrees')
    #
    #        period.Units = bounds_data.Units
    #
    #        if abs(bounds[-1, -1] - bounds[0, 0]) != period.array:
    #            if verbose:
    #                print(4)
    #            return False
    #
    #        self.period(period)
    #
    #        return True

    @_deprecated_kwarg_check("i")
    @_inplace_enabled(default=False)
    def roll(self, axis, shift, inplace=False, i=False):
        """Rolls the dimension coordinate along a cyclic axis."""
        c = _inplace_enabled_define_and_cleanup(self)

        size = c.size
        if size <= 1:
            return c

        shift %= size
        if not shift:
            # Null roll
            return c

        #        period = self._custom.get('period')
        period = c.period()

        if period is None:
            raise ValueError(
                f"Can't roll {self.__class__.__name__} when no period has "
                "been set"
            )

        direction = c.direction()

        centre = c._centre(period)

        if axis not in [0, -1]:
            raise ValueError(
                f"Can't roll axis {axis} when there is only one axis"
            )

        super(DimensionCoordinate, c).roll(axis, shift, inplace=True)

        c.dtype = numpy_result_type(c.dtype, period.dtype)

        b = c.get_bounds(None)
        bounds_data = c.get_bounds_data(None, _fill_value=False)

        if bounds_data is not None:
            b.dtype = numpy_result_type(bounds_data.dtype, period.dtype)
            bounds_data = b.get_data(None, _fill_value=False)

        if direction:
            # Increasing
            c[:shift] -= period
            if bounds_data is not None:
                b[:shift] -= period

            if c.data[0] <= centre - period:
                c += period
                if bounds_data is not None:
                    b += period
        else:
            # Decreasing
            c[:shift] += period
            if bounds_data is not None:
                b[:shift] += period

            if c.data[0] >= centre + period:
                c -= period
                if bounds_data is not None:
                    b -= period

        c._custom["direction"] = direction

        return c

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    @property
    def role(self):
        """Deprecated at version 3.0.0, use `construct_type` attribute
        instead."""
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, "role", "Use attribute 'construct_type' instead"
        )  # pragma: no cover



from numpy import empty as numpy_empty
from numpy import full as numpy_full
from numpy.ma import masked_all as numpy_ma_masked_all

from ..constants import masked as cf_masked
from ..functions import parse_indices
from . import abstract


class FilledArray(abstract.Array):
    """An underlying filled array."""

    def __init__(
        self,
        dtype=None,
        ndim=None,
        shape=None,
        size=None,
        fill_value=None,
        masked_all=False,
    ):
        """**Initialization**

            :Parameters:

                dtype : numpy.dtype
                    The numpy data type of the data array.

                ndim : int
                    Number of dimensions in the data array.

                shape : tuple
                    The data array's dimension sizes.

                size : int
                    Number of elements in the data array.

                fill_value : scalar, optional

        #        masked_all: `bool`

        """
        super().__init__(
            dtype=dtype,
            ndim=ndim,
            shape=shape,
            size=size,
            fill_value=fill_value,
        )

    def __getitem__(self, indices):
        """x.__getitem__(indices) <==> x[indices]

        Returns a numpy array.

        """
        if indices is Ellipsis:
            array_shape = self.shape
        else:
            array_shape = []
            for index in parse_indices(self.shape, indices):
                if isinstance(index, slice):
                    step = index.step
                    if step == 1:
                        array_shape.append(index.stop - index.start)
                    elif step == -1:
                        stop = index.stop
                        if stop is None:
                            array_shape.append(index.start + 1)
                        else:
                            array_shape.append(index.start - index.stop)
                    else:
                        stop = index.stop
                        if stop is None:
                            stop = -1

                        a, b = divmod(stop - index.start, step)
                        if b:
                            a += 1
                        array_shape.append(a)
                else:
                    array_shape.append(len(index))
        # --- End: if

        if self.fill_value() is cf_masked:
            return numpy_ma_masked_all(array_shape, dtype=self.dtype)
        elif self.fill_value() is not None:
            return numpy_full(
                array_shape, fill_value=self.fill_value(), dtype=self.dtype
            )
        else:
            return numpy_empty(array_shape, dtype=self.dtype)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def dtype(self):
        """Data-type of the data elements.

        **Examples:**

        >>> a.dtype
        dtype('float64')
        >>> print(type(a.dtype))
        <type 'numpy.dtype'>

        """
        return self._get_component("dtype")

    @property
    def ndim(self):
        """Number of array dimensions.

        **Examples:**

        >>> a.shape
        (73, 96)
        >>> a.ndim
        2
        >>> a.size
        7008

        >>> a.shape
        (1, 1, 1)
        >>> a.ndim
        3
        >>> a.size
        1

        >>> a.shape
        ()
        >>> a.ndim
        0
        >>> a.size
        1

        """
        return self._get_component("ndim")

    @property
    def shape(self):
        """Tuple of array dimension sizes.

        **Examples:**

        >>> a.shape
        (73, 96)
        >>> a.ndim
        2
        >>> a.size
        7008

        >>> a.shape
        (1, 1, 1)
        >>> a.ndim
        3
        >>> a.size
        1

        >>> a.shape
        ()
        >>> a.ndim
        0
        >>> a.size
        1

        """
        return self._get_component("shape")

    @property
    def size(self):
        """Number of elements in the array.

        **Examples:**

        >>> a.shape
        (73, 96)
        >>> a.size
        7008
        >>> a.ndim
        2

        >>> a.shape
        (1, 1, 1)
        >>> a.ndim
        3
        >>> a.size
        1

        >>> a.shape
        ()
        >>> a.ndim
        0
        >>> a.size
        1

        """
        return self._get_component("size")

    def fill_value(self):
        """Return the data array missing data value."""
        return self._get_component("fill_value")

    @property
    def array(self):
        """An independent numpy array containing the data."""
        return self[...]

    def reshape(self, newshape):
        """Give a new shape to the array."""
        new = self.copy()
        new.shape = newshape
        new.ndim = len(newshape)
        return new

    def resize(self, newshape):
        """Change the shape and size of the array in-place."""
        self.shape = newshape
        self.ndim = len(newshape)

    def view(self):
        """Return a view of the entire array."""
        return self[...]



import datetime
import re
import unittest

import numpy

import cf


class DomainTest(unittest.TestCase):
    d = cf.example_field(1).domain

    def setUp(self):
        # Disable log messages to silence expected warnings
        cf.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cf.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cf.log_level('DISABLE')

    def test_Domain__repr__str__dump(self):
        d = self.d

        _ = repr(d)
        _ = str(d)
        self.assertIsInstance(d.dump(display=False), str)

    def test_Domain__init__(self):
        cf.Domain(source="qwerty")

    def test_Domain_equals(self):
        d = self.d
        e = d.copy()

        self.assertTrue(d.equals(d, verbose=3))
        self.assertTrue(d.equals(e, verbose=3))
        self.assertTrue(e.equals(d, verbose=3))

    def test_Domain_flip(self):
        f = self.d.copy()

        kwargs = {
            axis: slice(None, None, -1) for axis in f.domain_axes(todict=True)
        }
        g = f.subspace(**kwargs)

        h = f.flip()
        self.assertTrue(h.equals(g, verbose=1))

        h = f.flip(["X", "Z", "Y"])
        self.assertTrue(h.equals(g, verbose=1))

        h = f.flip((re.compile("^atmos"), "grid_latitude", "grid_longitude"))
        self.assertTrue(h.equals(g, verbose=1))

        g = f.subspace(grid_longitude=slice(None, None, -1))
        self.assertIsNone(f.flip("X", inplace=True))
        self.assertTrue(f.equals(g, verbose=1))

    def test_Domain_properties(self):
        d = cf.Domain()

        d.set_property("long_name", "qwerty")

        self.assertEqual(d.properties(), {"long_name": "qwerty"})
        self.assertEqual(d.get_property("long_name"), "qwerty")
        self.assertEqual(d.del_property("long_name"), "qwerty")
        self.assertIsNone(d.get_property("long_name", None))
        self.assertIsNone(d.del_property("long_name", None))

        d.set_property("long_name", "qwerty")
        self.assertEqual(d.clear_properties(), {"long_name": "qwerty"})

        d.set_properties({"long_name": "qwerty"})
        d.set_properties({"foo": "bar"})
        self.assertEqual(d.properties(), {"long_name": "qwerty", "foo": "bar"})

    def test_Domain_creation_commands(self):
        for f in cf.example_fields():
            _ = f.domain.creation_commands()

        f = cf.example_field(1).domain

        for rd in (False, True):
            _ = f.creation_commands(representative_data=rd)

        for indent in (0, 4):
            _ = f.creation_commands(indent=indent)

        for s in (False, True):
            _ = f.creation_commands(string=s)

        for ns in ("cf", ""):
            _ = f.creation_commands(namespace=ns)

    def test_Domain_subspace(self):
        f = self.d.copy()

        x = f.dimension_coordinate("X")
        a = x.varray
        a[...] = numpy.arange(0, 360, 40)
        x.set_bounds(x.create_bounds())
        f.cyclic("X", iscyclic=True, period=360)

        f0 = f.copy()

        # wi (increasing)
        g = f.subspace(grid_longitude=cf.wi(50, 130))
        self.assertEqual(g.size, 20)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [80, 120]).all())

        g = f.subspace(grid_longitude=cf.wi(-90, 50))
        self.assertEqual(g.size, 40)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [-80, -40, 0, 40]).all())

        g = f.subspace(grid_longitude=cf.wi(310, 450))
        self.assertEqual(g.size, 40)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [-40, 0, 40, 80]).all())

        g = f.subspace(grid_longitude=cf.wi(310 - 1080, 450 - 1080))
        self.assertEqual(g.size, 40)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [-40, 0, 40, 80]).all())

        g = f.subspace(grid_longitude=cf.wi(310 + 720, 450 + 720))
        self.assertEqual(g.size, 40)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [-40, 0, 40, 80]).all())

        g = f.subspace(grid_longitude=cf.wi(-90, 370))
        self.assertEqual(g.size, 90)
        x = g.construct("grid_longitude").array
        self.assertTrue(
            (x == [-80, -40, 0, 40, 80, 120, 160, 200, 240.0]).all()
        )

        with self.assertRaises(IndexError):
            f.indices(grid_longitude=cf.wi(90, 100))

        # wi (decreasing)
        f.flip("X", inplace=True)

        g = f.subspace(grid_longitude=cf.wi(50, 130))
        self.assertEqual(g.size, 20)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [80, 120][::-1]).all())

        g = f.subspace(grid_longitude=cf.wi(-90, 50))
        self.assertEqual(g.size, 40)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [-80, -40, 0, 40][::-1]).all())

        g = f.subspace(grid_longitude=cf.wi(310, 450))
        self.assertEqual(g.size, 40)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [-40, 0, 40, 80][::-1]).all())

        g = f.subspace(grid_longitude=cf.wi(310 - 1080, 450 - 1080))
        self.assertEqual(g.size, 40)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [-40, 0, 40, 80][::-1]).all())

        g = f.subspace(grid_longitude=cf.wi(310 + 720, 450 + 720))
        self.assertEqual(g.size, 40)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [-40, 0, 40, 80][::-1]).all())

        with self.assertRaises(IndexError):
            f.indices(grid_longitude=cf.wi(90, 100))

        # wo
        f = f0.copy()

        g = f.subspace(grid_longitude=cf.wo(50, 130))
        self.assertEqual(g.size, 70)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [-200, -160, -120, -80, -40, 0, 40]).all())

        with self.assertRaises(IndexError):
            f.indices(grid_longitude=cf.wo(-90, 370))

        # set
        g = f.subspace(grid_longitude=cf.set([320, 40, 80, 99999]))
        self.assertEqual(g.size, 30)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [40, 80, 320]).all())

        g = f.subspace(grid_longitude=cf.lt(90))
        self.assertEqual(g.size, 30)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [0, 40, 80]).all())

        g = f.subspace(grid_longitude=cf.gt(90))
        self.assertEqual(g.size, 60)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [120, 160, 200, 240, 280, 320]).all())

        g = f.subspace(grid_longitude=cf.le(80))
        self.assertEqual(g.size, 30)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [0, 40, 80]).all())

        g = f.subspace(grid_longitude=cf.ge(80))
        self.assertEqual(g.size, 70)
        x = g.construct("grid_longitude").array
        self.assertTrue((x == [80, 120, 160, 200, 240, 280, 320]).all())

        # 2-d
        lon = f.auxiliary_coordinate("X")
        lon.data[...] = numpy.arange(60, 150).reshape(9, 10)

        lat = f.auxiliary_coordinate("Y")
        lat.data[...] = numpy.arange(-45, 45).reshape(10, 9)

        for mode in ("compress", "envelope"):
            g = f.subspace(mode, longitude=cf.wi(92, 134))
            size = 50
            self.assertEqual(g.size, size)

        for mode in (("compress",), ("envelope",)):
            g = f.subspace(*mode, longitude=cf.wi(72, 83) | cf.gt(118))
            if mode == ("envelope",):
                size = 80
            else:
                size = 60

            self.assertEqual(g.size, size, mode)

        for mode in ("compress", "envelope"):
            g = f.subspace(mode, grid_longitude=cf.contains(23.2))
            size = 10
            self.assertEqual(g.size, size)
            self.assertEqual(g.construct("grid_longitude").array, 40)

        for mode in ("compress", "envelope"):
            g = f.subspace(mode, grid_latitude=cf.contains(1))
            size = 9
            self.assertEqual(g.size, size)
            self.assertEqual(g.construct("grid_latitude").array, 0.88)

        for mode in ("compress", "envelope"):
            g = f.subspace(mode, longitude=cf.contains(83))
            size = 1
            self.assertEqual(g.size, size)

            self.assertEqual(g.construct("longitude").array, 83)

        # Calls that should fail
        with self.assertRaises(Exception):
            f.subspace(grid_longitude=cf.gt(23), X=cf.wi(92, 134))
        with self.assertRaises(Exception):
            f.subspace(grid_longitude=cf.gt(23), longitude=cf.wi(92, 134))
        with self.assertRaises(Exception):
            f.subspace(grid_latitude=cf.contains(-23.2))

    def test_Domain_transpose(self):
        f = cf.example_field(1)
        d = f.domain

        axes = [re.compile("^atmos"), "grid_latitude", "X"]

        g = f.transpose(axes, constructs=True)
        e = d.transpose(axes + ["T"])
        self.assertTrue(e.equals(g.domain))

        self.assertIsNone(e.transpose(axes + ["T"], inplace=True))

        with self.assertRaises(ValueError):
            d.transpose(["X", "Y"])

        with self.assertRaises(ValueError):
            d.transpose(["X", "Y", 1])

        with self.assertRaises(ValueError):
            d.transpose([2, 1])

        with self.assertRaises(ValueError):
            d.transpose(["Y", "Z"])

        with self.assertRaises(ValueError):
            d.transpose(["Y", "Y", "Z"])

        with self.assertRaises(ValueError):
            d.transpose(["Y", "X", "Z", "Y"])

        with self.assertRaises(ValueError):
            d.transpose(["Y", "X", "Z", 1])


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print("")
    unittest.main(verbosity=2)



from ..functions import _DEPRECATION_ERROR_ARG


class FieldDomainList:
    """A sequence of domain constructs corresponding to fields."""

    def select_by_construct(self, *identities, OR=False, **conditions):
        """Select elements by their metadata constructs.

        To find the inverse of the selection, use a list comprehension
        with the `!match_by_construct` method of the construct
        elements. For example, to select all constructs that do *not*
        have a "latitude" metadata construct:

           >>> gl = cf.{{class}}(
           ...     f for f in fl if not f.match_by_construct('latitude')
           ... )

        .. note:: The API changed at version 3.1.0

        .. versionadded:: 3.0.0

        .. seealso: `select`, `__call__`, `select_by_units`,
                    `select_by_naxes`, `select_by_rank`,
                    `select_by_property`

        :Parameters:

            identities: optional
                Identify metadata constructs that have an identity,
                defined by their `!identities` methods, that matches
                any of the given values.

                If no identities nor conditions (see the *conditions*
                parameter) are provided then all constructs are
                selected.

                {{value match}}

                {{displayed identity}}

                If a cell method construct identity is given (such as
                ``'method:mean'``) then it will only be compared with
                the most recently applied cell method operation.

                Alternatively, one or more cell method constructs may
                be identified in a single string with a CF-netCDF cell
                methods-like syntax for describing both the collapse
                dimensions, the collapse method, and any cell method
                construct qualifiers. If N cell methods are described
                in this way then they will collectively identify the N
                most recently applied cell method operations. For
                example, ``'T: maximum within years T: mean over
                years'`` will be compared with the most two most
                recently applied cell method operations.

                *Parameter example:*
                  ``'latitude'``

                *Parameter example:*
                  ``'T'``

                *Parameter example:*
                  ``'long_name=Cell Area'``

                *Parameter example:*
                  ``'cellmeasure1'``

                *Parameter example:*
                  ``'measure:area'``

                *Parameter example:*
                  ``cf.eq('time')'``

                *Parameter example:*
                  ``re.compile('^lat')``

                *Parameter example:*
                  ``'domainancillary2', 'longitude'``

                *Parameter example:*
                  ``'area: mean T: maximum'``

                *Parameter example:*
                  ``'grid_latitude', 'area: mean T: maximum'``

            conditions: optional
                Identify the metadata constructs that have any of the
                given identities or construct keys, and whose data satisfy
                conditions.

                A construct identity or construct key (defined in the
                same way as by the *identities* parameter) is given as
                a keyword name and a condition on its data is given as
                the keyword value.

                The condition is satisfied if any of its data values
                equals the value provided.

                If no conditions nor identities (see the *identities*
                parameter) are provided then all list elements are
                selected.

                *Parameter example:*
                  ``longitude=180.0``

                *Parameter example:*
                  ``time=cf.dt('1959-12-16')``

                *Parameter example:*
                  ``latitude=cf.ge(0)``

                *Parameter example:*
                  ``latitude=cf.ge(0), air_pressure=500``

                *Parameter example:*
                  ``**{'latitude': cf.ge(0), 'long_name=soil_level': 4}``

            OR: `bool`, optional
                If True then return `True` if at least one metadata
                construct matches at least one of the criteria given
                by the *identities* or *conditions* arguments. By
                default `True` is only returned if the field
                constructs matches each of the given criteria.

            mode: deprecated at version 3.1.0
                Use the *OR* parameter instead.

            constructs: deprecated at version 3.1.0

        :Returns:

            `bool`
                The matching field constructs.

        **Examples**

            TODO

        """
        if identities:
            if identities[0] == "or":
                _DEPRECATION_ERROR_ARG(
                    self,
                    "select_by_construct",
                    "or",
                    message="Use 'OR=True' instead.",
                    version="3.1.0",
                )  # pragma: no cover

            if identities[0] == "and":
                _DEPRECATION_ERROR_ARG(
                    self,
                    "select_by_construct",
                    "and",
                    message="Use 'OR=False' instead.",
                    version="3.1.0",
                )  # pragma: no cover

        return type(self)(
            f
            for f in self
            if f.match_by_construct(*identities, OR=OR, **conditions)
        )

    def select_by_ncvar(self, *ncvars):
        """Select list elements by netCDF variable name.

        To find the inverse of the selection, use a list comprehension
        with the `!match_by_ncvar` method of the construct
        elements. For example, to select all constructs which do *not*
        have a netCDF name of 'tas':

           >>> gl = cf.{{class}}(
           ...     f for f in fl if not f.match_by_ncvar('tas')
           ... )

        .. versionadded:: 3.0.0

        .. seealso:: `select`, `select_by_identity`, `select_by_property`,

        :Parameters:

            ncvars: optional
                Select constructs from the list. May be one or more
                netCDF names of constructs.

                A construct is selected if it matches any of the given
                names.

                A netCDF variable name is specified by a string (e.g.
                ``'tas'``, etc.); a `Query` object

                (e.g. ``cf.eq('tas')``); or a compiled regular
                expression (e.g. ``re.compile('^air_')``) that selects
                the constructs whose netCDF variable names match via
                `re.search`.

                If no netCDF variable names are provided then all are
                selected.

        :Returns:

            `{{class}}`
                The matching constructs.

        **Examples**

        >>> fl = cf.{{class}}([cf.example_field(0), cf.example_field(1)])
        >>> fl
        [<CF Field: specific_humidity(latitude(5), longitude(8)) 1>,
         <CF Field: air_temperature(atmosphere_hybrid_height_coordinate(1), grid_latitude(10), grid_longitude(9)) K>]
        >>> f[0].nc_get_variable()
        'humidity'
        >>> f[1].nc_get_variable()
        'temp'

        >>> fl.select_by_ncvar('humidity')
        [<CF Field: specific_humidity(cf_role=timeseries_id(4), ncdim%timeseries(9))>]
        >>> fl.select_by_ncvar('humidity', 'temp')
        [<CF Field: specific_humidity(cf_role=timeseries_id(4), ncdim%timeseries(9))>,
         <CF Field: air_temperature(cf_role=timeseries_id(4), ncdim%timeseries(9)) Celsius>]
        >>> fl.select_by_ncvar()
        [<CF Field: specific_humidity(cf_role=timeseries_id(4), ncdim%timeseries(9))>,
         <CF Field: air_temperature(cf_role=timeseries_id(4), ncdim%timeseries(9)) Celsius>]

        >>> import re
        >>> fl.select_by_ncvar(re.compile('^hum'))
        [<CF Field: specific_humidity(cf_role=timeseries_id(4), ncdim%timeseries(9))>]

        """
        return type(self)(f for f in self if f.match_by_ncvar(*ncvars))

    def select_by_property(self, *mode, **properties):
        """Select list elements by property.

        To find the inverse of the selection, use a list comprehension
        with the `!match_by_property` method of the construct
        elements. For example, to select all constructs which do *not*
        have a long_name property of "Pressure":

           >>> gl = cf.{{class}}(
           ...     f for f in fl if not f.match_by_property(long_name='Pressure')
           ... )

        .. versionadded:: 3.0.0

        .. seealso:: `select`, `select_by_identity`, `select_by_ncvar`

        :Parameters:

            mode: optional
                Define the behaviour when multiple properties are
                provided.

                By default (or if the *mode* parameter is ``'and'``) a
                construct is selected if it matches all of the given
                properties, but if the *mode* parameter is ``'or'``
                then a construct will be selected when at least one of
                its properties matches.

            properties: optional
                Select the constructs with the given properties.

                A property value is given by a keyword parameter of
                the property name. The value may be a scalar or vector
                (e.g. ``'air_temperature'``, ``4``, ``['foo',
                'bar']``); or a compiled regular expression
                (e.g. ``re.compile('^ocean')``), for which all
                constructs whose methods match (via `re.search`) are
                selected.

                By default a construct is selected if it matches all
                of the given properties, but it may alternatively be
                selected when at least one of its properties matches
                (see the *mode* positional parameter).

        :Returns:

            `{{class}}`
                The matching constructs.

        **Examples**

        See `cf.{{class}}.select_by_identity`

        """
        return type(self)(
            f for f in self if f.match_by_property(*mode, **properties)
        )

    def select_by_rank(self, *ranks):
        """Select list elements by the number of domain axis constructs.

        .. versionadded:: 3.0.0

        .. seealso: `select`, `__call__`, `select_by_units`,
                    `select_by_naxes`, `select_by_construct`,
                    `select_by_property`, `cf.Field.match_by_identity`

        :Parameters:

            ranks: optional
                Define conditions on the number of domain axis constructs.

                A condition is one of:

                  * `int`
                  * a `Query` object

                The condition is satisfied if the number of domain
                axis constructs equals the condition value.

                *Parameter example:*
                  To see if the field construct has 4 domain axis
                  constructs: ``4``

                *Parameter example:*
                  To see if the field construct has at least 3 domain
                  axis constructs: ``cf.ge(3)``

        :Returns:

            `bool`
                The matching field constructs.

        **Examples**

            TODO

        """

        return type(self)(f for f in self if f.match_by_rank(*ranks))



import abc
from functools import reduce
from operator import mul
from sys import getrefcount

from ...functions import inspect as cf_inspect


class CompressedSubarray(abc.ABC):
    """Abstract base class for a compressed sub-array container."""

    def __init__(self, array, shape, compression):
        """**Initialization**

        :Parameters:

            array:

            shape: `tuple`
                The shape of the uncompressed array

            compression: `dict`

        """
        # DO NOT CHANGE IN PLACE
        self.array = array

        # DO NOT CHANGE IN PLACE
        self.compression = compression

        # DO NOT CHANGE IN PLACE
        self.shape = tuple(shape)

        # DO NOT CHANGE IN PLACE
        self.ndim = len(shape)

        # DO NOT CHANGE IN PLACE
        self.size = reduce(mul, shape, 1)

    @abc.abstractmethod
    def __getitem__(self, indices):
        """x.__getitem__(indices) <==> x[indices]

        Returns a numpy array.

        """
        raise NotImplementedError()  # pragma: no cover

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        array = self.array
        shape = str(array.shape)
        shape = shape.replace(",)", ")")

        return "<CF {}{}: {}>".format(
            self.__class__.__name__, shape, str(array)
        )

    @property
    def dtype(self):
        return self.array.dtype

    @property
    def file(self):
        """The file on disk which contains the compressed array, or
        `None` of the array is in memory.

        **Examples:**

         >>> self.file
         '/home/foo/bar.nc'

        """
        return getattr(self.array, "file", None)

    def close(self):
        """Close all referenced open files.

        :Returns:

            `None`

        **Examples:**

        >>> f.close()

        """
        if self.on_disk():
            self.array.close()

    def copy(self):
        """Replace the abstract base class with a deep copy."""
        C = self.__class__
        new = C.__new__(C)
        new.__dict__ = self.__dict__.copy()
        return new

    def inspect(self):
        """Inspect the object for debugging.

        .. seealso:: `cf.inspect`

        :Returns:

            `None`

        """
        print(cf_inspect(self))

    def on_disk(self):
        """True if and only if the compressed array is on disk as
        opposed to in memory.

        **Examples:**

        >>> a.on_disk()
        True

        """
        return not hasattr(self.array, "__array_interface__")

    def unique(self):
        """True if there is only one permanent reference to the array
        instance."""
        # Note, from the Python docs for sys.getrefcount:
        # "The count returned is generally one higher than you might expect,
        # because it includes the (temporary) reference as an argument to
        # getrefcount", hence <= 2 to test for uniqueness rather than <= 1.
        return getrefcount(self.array) <= 2



import json
import random
from os.path import dirname, isfile
from string import hexdigits

import cfdm
import numpy
from netCDF4 import Dataset as netCDF4_Dataset

from ... import Bounds, Coordinate, DomainAncillary

# TODO: is it OK to import from here? Maybe best move these to '...functions'?
from ...data.functions import _close_netcdf_file, _file_to_Dataset
from ...functions import relpath


class NetCDFWrite(cfdm.read_write.netcdf.NetCDFWrite):
    """A container for writing Fields to a netCDF dataset."""

    def file_close(self, filename):
        """Close the netCDF file that has been written.

        :Returns:

            `None`

        """
        _close_netcdf_file(filename)

    def file_open(self, filename, mode, fmt, fields):
        """Open the netCDF file for writing.

        :Parameters:

            filename: `str`
                As for the *filename* parameter for initialising a
                `netCDF.Dataset` instance.

            mode: `str`
                As for the *mode* parameter for initialising a
                `netCDF.Dataset` instance.

            fmt: `str`
                As for the *format* parameter for initialising a
                `netCDF.Dataset` instance.

            fields: sequence of `Field`
                The field constructs to be written.

        :Returns:

            `netCDF.Dataset`
                A `netCDF4.Dataset` object for the file.

        """
        # I.e. if on either file IO iteration for append mode:
        if self.write_vars["dry_run"] or self.write_vars["post_dry_run"]:
            if not isfile(filename):
                nc = netCDF4_Dataset(filename, "w", format=fmt)
                nc.close()
            elif filename in _file_to_Dataset:
                _close_netcdf_file(filename)

        nc = super().file_open(filename, mode, fmt, fields)
        _file_to_Dataset[filename] = nc

        return nc

    def _write_as_cfa(self, cfvar):
        """True if the variable should be written as a CFA variable.

        .. versionadded:: 3.0.0

        """
        if not self.write_vars["cfa"]:
            return False

        data = self.implementation.get_data(cfvar, None)
        if data is None:
            return False

        if data.in_memory:
            return False

        if data.size == 1:
            return False

        if isinstance(cfvar, (Coordinate, DomainAncillary)):
            return cfvar.ndim > 1

        if isinstance(cfvar, Bounds):
            return cfvar.ndim > 2

        return True

    def _customize_createVariable(self, cfvar, kwargs):
        """Customise keyword arguments for
        `netCDF4.Dataset.createVariable`.

        .. versionadded:: 3.0.0

        :Parameters:

            cfvar: cf instance that contains data

            kwargs: `dict`

        :Returns:

            `dict`
                Dictionary of keyword arguments to be passed to
                `netCDF4.Dataset.createVariable`.

        """
        kwargs = super()._customize_createVariable(cfvar, kwargs)

        if self._write_as_cfa(cfvar):
            kwargs["dimensions"] = ()
            kwargs["chunksizes"] = None

        return kwargs

    def _write_data(
        self,
        data,
        cfvar,
        ncvar,
        ncdimensions,
        unset_values=(),
        compressed=False,
        attributes={},
    ):
        """Write a Data object.

        .. versionadded:: 3.0.0

        :Parameters:

            data: `Data`

            cfvar: cf instance

            ncvar: `str`

            ncdimensions: `tuple` of `str`

            unset_values: sequence of numbers

        """
        g = self.write_vars

        if self._write_as_cfa(cfvar):
            self._write_cfa_data(ncvar, ncdimensions, data, cfvar)
            return

        # Still here?
        if compressed:
            # --------------------------------------------------------
            # Write data in its compressed form
            # --------------------------------------------------------
            data = data.source().source()

        warned_valid = False

        config = data.partition_configuration(readonly=True)

        for partition in data.partitions.flat:
            partition.open(config)
            array = partition.array

            # Convert data type
            new_dtype = g["datatype"].get(array.dtype)
            if new_dtype is not None:
                array = array.astype(new_dtype)

            # Check that the array doesn't contain any elements
            # which are equal to any of the missing data values
            if unset_values:
                if partition.masked:
                    temp_array = array.compressed()
                else:
                    temp_array = array

                if numpy.intersect1d(unset_values, temp_array).size:
                    raise ValueError(
                        "ERROR: Can't write field when array has _FillValue or"
                        f" missing_value at unmasked point: {ncvar!r}"
                    )
            # --- End: if

            if (
                g["fmt"] == "NETCDF4"
                and array.dtype.kind in "SU"
                and numpy.ma.isMA(array)
            ):
                # VLEN variables can not be assigned to by masked arrays
                # https://github.com/Unidata/netcdf4-python/pull/465
                array = array.filled("")

            if not warned_valid and g["warn_valid"]:
                # Check for out-of-range values
                warned_valid = self._check_valid(cfvar, array, attributes)

            # Copy the array into the netCDF variable
            g["nc"][ncvar][partition.indices] = array

            partition.close()

    def _write_dimension_coordinate(
        self, f, key, coord, ncdim=None, coordinates=None
    ):
        """Write a coordinate variable and its bound variable to the
        file.

        This also writes a new netCDF dimension to the file and, if
        required, a new netCDF dimension for the bounds.

        .. versionadded:: 3.0.0

        :Parameters:

            f: Field construct

            key: `str`

            coord: Dimension coordinate construct

            ncdim: `str` or `None`
                The name of the netCDF dimension for this dimension
                coordinate construct, including any groups structure. Note
                that the group structure may be different to the
                corodinate variable, and the basename.

                .. versionadded:: 3.6.0

            coordinates: `list`
               This list may get updated in-place.

                .. versionadded:: 3.7.0

        :Returns:

            `str`
                The netCDF name of the dimension coordinate.

        """
        coord = self._change_reference_datetime(coord)

        return super()._write_dimension_coordinate(
            f, key, coord, ncdim=ncdim, coordinates=coordinates
        )

    def _write_scalar_coordinate(
        self, f, key, coord_1d, axis, coordinates, extra=None
    ):
        """Write a scalar coordinate and its bounds to the netCDF file.

        It is assumed that the input coordinate has size 1, but this is
        not checked.

        If an equal scalar coordinate has already been written to the file
        then the input coordinate is not written.

        .. versionadded:: 3.0.0

        :Parameters:

            f: Field construct

            key: `str`
                The coordinate construct key

            coord_1d: Coordinate construct

            axis: `str`
                The field's axis identifier for the scalar coordinate.

            coordinates: `list`

        :Returns:

            coordinates: `list`
                The updated list of netCDF auxiliary coordinate names.

        """
        # Unsafe to set mutable '{}' as default in the func signature.
        if extra is None:  # distinguish from falsy '{}'
            extra = {}

        coord_1d = self._change_reference_datetime(coord_1d)

        return super()._write_scalar_coordinate(
            f, key, coord_1d, axis, coordinates, extra=extra
        )

    def _write_auxiliary_coordinate(self, f, key, coord, coordinates):
        """Write auxiliary coordinates and bounds to the netCDF file.

        If an equal auxiliary coordinate has already been written to the
        file then the input coordinate is not written.

        .. versionadded:: 3.0.0

        :Parameters:

            f: Field construct

            key: `str`

            coord: Coordinate construct

            coordinates: `list`

        :Returns:

            coordinates: `list`
                The list of netCDF auxiliary coordinate names updated in
                place.

        **Examples**

        >>> coordinates = _write_auxiliary_coordinate(f, 'aux2', coordinates)

        """
        coord = self._change_reference_datetime(coord)

        return super()._write_auxiliary_coordinate(f, key, coord, coordinates)

    def _change_reference_datetime(self, coord):
        """Change the units of a reference date-time value.

        .. versionadded:: 3.0.0

        :Parameters:

            coord: Coordinate instance

        :Returns:

                The coordinate construct with changed units.

        """
        reference_datetime = self.write_vars.get("reference_datetime")
        if not reference_datetime or not coord.Units.isreftime:
            return coord

        coord2 = coord.copy()
        try:
            coord2.reference_datetime = reference_datetime
        except ValueError:
            raise ValueError(
                "Can't override coordinate reference date-time "
                f"{coord.reference_datetime!r} with {reference_datetime!r}"
            )
        else:
            return coord2

    def _write_cfa_data(self, ncvar, ncdimensions, data, cfvar):
        """Write a CFA variable to the netCDF file.

        Any CFA private variables required will be autmatically created
        and written to the file.

        .. versionadded:: 3.0.0

        :Parameters:

            ncvar: `str`
                The netCDF name for the variable.

            ncdimensions: sequence of `str`

            netcdf_attrs: `dict`

            data: `Data`

        :Returns:

            `None`

        """
        g = self.write_vars

        netcdf_attrs = {
            "cf_role": "cfa_variable",
            "cfa_dimensions": " ".join(ncdimensions),
        }

        # Create a dictionary representation of the data object
        data = data.copy()
        axis_map = {}
        for axis0, axis1 in zip(data._axes, ncdimensions):
            axis_map[axis0] = axis1

        data._change_axis_names(axis_map)
        data._move_flip_to_partitions()

        cfa_array = data.dumpd()

        # Modify the dictionary so that it is suitable for JSON
        # serialization
        del cfa_array["_axes"]
        del cfa_array["shape"]
        del cfa_array["Units"]
        del cfa_array["dtype"]
        cfa_array.pop("_cyclic", None)
        cfa_array.pop("_fill", None)
        cfa_array.pop("fill_value", None)

        pmshape = cfa_array.pop("_pmshape", None)
        if pmshape:
            cfa_array["pmshape"] = pmshape

        pmaxes = cfa_array.pop("_pmaxes", None)
        if pmaxes:
            cfa_array["pmdimensions"] = pmaxes

        config = data.partition_configuration(readonly=True)

        base = g["cfa_options"].get("base", None)
        if base is not None:
            if base == "":
                cfa_array["base"] = dirname(self.write_vars["filename"])
            else:
                cfa_array["base"] = base

        convert_dtype = g["datatype"]

        for attrs in cfa_array["Partitions"]:
            fmt = attrs.get("format", None)

            if fmt is None:
                # --------------------------------------------------------
                # This partition has an internal sub-array. This could be
                # a numpy array or a temporary FileArray object.
                # --------------------------------------------------------
                index = attrs.get("index", ())
                if len(index) == 1:
                    index = index[0]
                else:
                    index = tuple(index)

                partition = data.partitions.matrix.item(index)

                partition.open(config)
                array = partition.array

                # Convert data type
                new_dtype = convert_dtype.get(array.dtype, None)
                if new_dtype is not None:
                    array = array.astype(new_dtype)

                shape = array.shape
                ncdim_strlen = []
                if array.dtype.kind == "S":
                    # This is an array of strings
                    strlen = array.dtype.itemsize
                    if strlen > 1:
                        # Convert to an array of characters
                        array = self._character_array(array)

                        # Get the netCDF dimension for the string length
                        ncdim_strlen = [
                            self._string_length_dimension(strlen, g=None)
                        ]
                # --- End: if

                # Create a name for the netCDF variable to contain the array
                p_ncvar = "cfa_" + self._random_hex_string()
                while p_ncvar in g["ncvar_names"]:
                    p_ncvar = "cfa_" + self._random_hex_string()

                g["ncvar_names"].add(p_ncvar)

                # Get the private CFA netCDF dimensions for the array.
                cfa_dimensions = [
                    self._netcdf_name(
                        f"cfa{size}", dimsize=size, role="cfa_private"
                    )
                    for size in array.shape
                ]

                for (
                    ncdim,
                    size,
                ) in zip(cfa_dimensions, array.shape):
                    if ncdim not in g["ncdim_to_size"]:
                        # This cfa private dimension needs creating
                        g["ncdim_to_size"][ncdim] = size
                        g["netcdf"].createDimension(ncdim, size)

                # Create the private CFA variable and write the array to it
                #                v = g['netcdf'].createVariable(p_ncvar, self._datatype(array),
                #                                               cfa_dimensions + ncdim_strlen,
                # #                                              fill_value=fill_value,
                #                                               fill_value=False,
                #                                               least_significant_digit=None,
                #                                               endian=g['endian'],
                #                                               **g['netcdf_compression'])

                kwargs = {
                    "varname": p_ncvar,
                    "datatype": self._datatype(array),
                    "dimensions": cfa_dimensions + ncdim_strlen,
                    "fill_value": None,  # False,
                    "least_significant_digit": None,
                    "endian": g["endian"],
                }
                kwargs.update(g["netcdf_compression"])

                self._createVariable(**kwargs)

                self._write_attributes(
                    parent=None,
                    ncvar=p_ncvar,
                    extra={"cf_role": "cfa_private"},
                )

                g["nc"][p_ncvar][...] = array

                # Update the attrs dictionary.
                #
                # Note that we don't need to set 'part', 'dtype', 'units',
                # 'calendar', 'dimensions' and 'reverse' since the
                # partition's in-memory data array always matches up with
                # the master data array.
                attrs["subarray"] = {"shape": shape, "ncvar": p_ncvar}

            else:
                # --------------------------------------------------------
                # This partition has an external sub-array
                # --------------------------------------------------------
                # PUNITS, PCALENDAR: Change from Units object to netCDF
                #                    string(s)
                units = attrs.pop("Units", None)
                if units is not None:
                    attrs["punits"] = units.units
                    if hasattr(units, "calendar"):
                        attrs["pcalendar"] = units.calendar

                # PDIMENSIONS:
                p_axes = attrs.pop("axes", None)
                if p_axes is not None:
                    attrs["pdimensions"] = p_axes

                # REVERSE
                p_flip = attrs.pop("flip", None)
                if p_flip:
                    attrs["reverse"] = p_flip

                # DTYPE: Change from numpy.dtype object to netCDF string
                dtype = attrs["subarray"].pop("dtype", None)
                if dtype is not None:
                    if dtype.kind != "S":
                        attrs["subarray"][
                            "dtype"
                        ] = self._convert_to_netCDF_datatype(dtype)

                # FORMAT:
                sfmt = attrs.pop("format", None)
                if sfmt is not None:
                    attrs["subarray"]["format"] = sfmt
            # --- End: if

            # LOCATION: Change from python to CFA indexing (i.e. range
            #           includes the final index)
            attrs["location"] = [(x[0], x[1] - 1) for x in attrs["location"]]

            # PART: Change from python to to CFA indexing (i.e. slice
            #       range includes the final index)
            part = attrs.get("part", None)
            if part:
                p = []
                for x, size in zip(part, attrs["subarray"]["shape"]):
                    if isinstance(x, slice):
                        x = x.indices(size)
                        if x[2] > 0:
                            p.append([x[0], x[1] - 1, x[2]])
                        elif x[1] == -1:
                            p.append([x[0], 0, x[2]])
                        else:
                            p.append([x[0], x[1] + 1, x[2]])
                    else:
                        p.append(tuple(x))
                # --- End: for
                attrs["part"] = str(p)
            # --- End: if

            if "base" in cfa_array and "file" in attrs["subarray"]:
                # Make the file name relative to base
                attrs["subarray"]["file"] = relpath(
                    attrs["subarray"]["file"], cfa_array["base"]
                )
        # --- End: for

        if base == "":
            cfa_array["base"] = base

        # Add the description (as a JSON string) of the partition array to
        # the netcdf attributes.
        netcdf_attrs["cfa_array"] = json.dumps(
            cfa_array, default=self._convert_to_builtin_type
        )

        # Write the netCDF attributes to the file
        self._write_attributes(parent=None, ncvar=ncvar, extra=netcdf_attrs)

    def _random_hex_string(self, size=10):
        """Return a random hexadecimal string with the given number of
        characters.

        .. versionadded:: 3.0.0

        :Parameters:

            size: `int`, optional
                The number of characters in the generated string.

        :Returns:

            `str`
                The hexadecimal string.

        **Examples**

        >>> _random_hex_string()
        'C3eECbBBcf'
        >>> _random_hex_string(6)
        '7a4acc'

        """
        return "".join(random.choice(hexdigits) for i in range(size))

    def _convert_to_builtin_type(self, x):
        """Convert a non-JSON-encodable object to a JSON-encodable
        built-in type.

        Possible conversions are:

        ==============  =============  ======================================
        Input object    Output object  numpy data types covered
        ==============  =============  ======================================
        numpy.bool_     bool           bool
        numpy.integer   int            int, int8, int16, int32, int64, uint8,
                                       uint16, uint32, uint64
        numpy.floating  float          float, float16, float32, float64
        ==============  =============  ======================================

        .. versionadded:: 3.0.0

        :Parameters:

            x:

        :Returns:

            'int' or `float` or `bool`

        **Examples**

        >>> type(_convert_to_builtin_type(numpy.bool_(True)))
        bool
        >>> type(_convert_to_builtin_type(numpy.array([1.0])[0]))
        double
        >>> type(_convert_to_builtin_type(numpy.array([2])[0]))
        int

        """
        if isinstance(x, numpy.bool_):
            return bool(x)

        if isinstance(x, numpy.integer):
            return int(x)

        if isinstance(x, numpy.floating):
            return float(x)

        raise TypeError(
            f"{type(x)!r} object can't be converted to a JSON serializable "
            f"type: {x!r}"
        )



import cfdm


class RaggedIndexedContiguousArray(cfdm.RaggedIndexedContiguousArray):
    """An underlying indexed contiguous ragged array.

    A collection of features, each of which is sequence of (vertical)
    profiles, stored using an indexed contiguous ragged array combines
    all feature elements along a single dimension (the "sample
    dimension") such that a contiguous ragged array representation is
    used for each profile and the indexed ragged array representation
    to organise the profiles into timeseries.

    The information needed to uncompress the data is stored in a
    "count variable" that gives the size of each profile; and in a
    "index variable" that specifies the feature that each profile
    belongs to.

    .. versionadded:: 3.0.0

    """



import cfdm

from . import mixin


class PartNodeCountProperties(mixin.Properties, cfdm.PartNodeCountProperties):
    """Properties for a netCDF part node count variable.

    **NetCDF interface**

    The netCDF part node count variable name may be accessed with the
    `nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
    `nc_has_variable` methods.

    The name of the netCDF dimension spanned by the netCDF part node
    count variable's data may be accessed with the `nc_set_dimension`,
    `nc_get_dimension`, `nc_del_dimension` and `nc_has_dimension`
    methods.

    .. versionadded:: 3.2.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)
